#!/usr/bin/env python
#
# Copyright (C) 2007 Andrew Resch <andrewresch@gmail.com>
# Copyright (C) 2009 Damien Churchill <damoxc@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import glob
import os
import platform
import sys
from distutils.command.build import build as _build
from distutils.command.clean import clean as _clean
from distutils.command.install_data import install_data as _install_data
from shutil import rmtree, which

from setuptools import Command, find_packages, setup
from setuptools.command.test import test as _test

import msgfmt
from version import get_version

try:
    from sphinx.setup_command import BuildDoc
except ImportError:

    class BuildDoc:
        pass


def windows_check():
    return platform.system() in ('Windows', 'Microsoft')


def osx_check():
    return platform.system() == 'Darwin'


desktop_data = 'deluge/ui/data/share/applications/deluge.desktop'
appdata_data = 'deluge/ui/data/share/appdata/deluge.appdata.xml'

# Variables for setuptools.setup
_package_data = {}
_exclude_package_data = {}
_entry_points = {'console_scripts': [], 'gui_scripts': [], 'deluge.ui': []}
_data_files = []
_version = get_version(prefix='deluge-', suffix='.dev0')


class PyTest(_test):
    def initialize_options(self):
        _test.initialize_options(self)
        self.pytest_args = []

    def finalize_options(self):
        _test.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest

        errcode = pytest.main(self.test_args)
        sys.exit(errcode)


class CleanDocs(Command):
    description = 'Clean the documentation build and module rst files'
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        docs_build = 'docs/build'
        print(f'Deleting {docs_build}')
        try:
            rmtree(docs_build)
        except OSError:
            pass

        for module in glob.glob('docs/source/modules/deluge*.rst'):
            os.remove(module)


class BuildWebUI(Command):
    description = 'Minify WebUI files'
    user_options = []

    JS_DIR = os.path.join('deluge', 'ui', 'web', 'js')
    JS_SRC_DIRS = ('deluge-all', os.path.join('extjs', 'ext-extensions'))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        js_basedir = os.path.join(os.path.dirname(__file__), self.JS_DIR)

        try:
            from minify_web_js import minify_js_dir

            import_error = ''
        except ImportError as err:
            import_error = err

        for js_src_dir in self.JS_SRC_DIRS:
            source_dir = os.path.join(js_basedir, js_src_dir)
            try:
                minify_js_dir(source_dir)
            except NameError:
                js_file = source_dir + '.js'
                if os.path.isfile(js_file):
                    print(
                        'Unable to minify but found existing minified: {}'.format(
                            js_file
                        )
                    )
                else:
                    # Unable to minify and no existing minified file found so exiting.
                    print('Import error: %s' % import_error)
                    sys.exit(1)

        # Create the gettext.js file for translations.
        try:
            from gen_web_gettext import create_gettext_js
        except ImportError:
            pass
        else:
            deluge_all_path = os.path.join(js_basedir, self.JS_SRC_DIRS[0])
            print('Creating WebUI translation file: %s/gettext.js' % deluge_all_path)
            create_gettext_js(deluge_all_path)


class CleanWebUI(Command):
    description = 'Clean the documentation build and rst files'
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        js_basedir = os.path.join(os.path.dirname(__file__), BuildWebUI.JS_DIR)

        # Remove files generated by minify script.
        for js_src_dir in BuildWebUI.JS_SRC_DIRS:
            for file_type in ('.js', '-debug.js'):
                js_file = os.path.join(js_basedir, js_src_dir + file_type)
                print(f'Deleting {js_file}')
                try:
                    os.remove(js_file)
                except OSError:
                    pass

        # Remove generated gettext.js
        js_file = os.path.join(js_basedir, 'gettext.js')
        print(f'Deleting {js_file}')
        try:
            os.remove(js_file)
        except OSError:
            pass


class BuildTranslations(Command):
    description = 'Compile .po files into .mo files & create .desktop file'

    user_options = [
        ('build-lib', None, 'lib build folder'),
        ('develop', 'D', 'Compile translations in develop mode (deluge/i18n)'),
    ]
    boolean_options = ['develop']

    def initialize_options(self):
        self.build_lib = None
        self.develop = False

    def finalize_options(self):
        self.set_undefined_options('build', ('build_lib', 'build_lib'))

    def run(self):
        po_dir = os.path.join(os.path.dirname(__file__), 'deluge', 'i18n')

        if self.develop:
            basedir = po_dir
        else:
            basedir = os.path.join(self.build_lib, 'deluge', 'i18n')

        intltool_merge = 'intltool-merge'
        if not windows_check() and which(intltool_merge):
            intltool_merge_opts = '--utf8 --quiet'
            for data_file in (desktop_data, appdata_data):
                # creates the translated file from .in file.
                in_file = data_file + '.in'
                if 'xml' in data_file:
                    intltool_merge_opts += ' --xml-style'
                elif 'desktop' in data_file:
                    intltool_merge_opts += ' --desktop-style'

                print('Creating file: %s' % data_file)
                os.system(
                    'C_ALL=C '
                    + '%s '
                    * 5
                    % (intltool_merge, intltool_merge_opts, po_dir, in_file, data_file)
                )

        print('Compiling po files from %s...' % po_dir)
        for path, names, filenames in os.walk(po_dir):
            for f in filenames:
                upto_date = False
                if f.endswith('.po'):
                    lang = f[: len(f) - 3]
                    src = os.path.join(path, f)
                    dest_path = os.path.join(basedir, lang, 'LC_MESSAGES')
                    dest = os.path.join(dest_path, 'deluge.mo')
                    if not os.path.exists(dest_path):
                        os.makedirs(dest_path)
                    if not os.path.exists(dest):
                        sys.stdout.write('%s, ' % lang)
                        sys.stdout.flush()
                        msgfmt.make(src, dest)
                    else:
                        src_mtime = os.stat(src)[8]
                        dest_mtime = os.stat(dest)[8]
                        if src_mtime > dest_mtime:
                            sys.stdout.write('%s, ' % lang)
                            sys.stdout.flush()
                            msgfmt.make(src, dest)
                        else:
                            upto_date = True

        if upto_date:
            sys.stdout.write(' po files already up to date.  ')
        sys.stdout.write('\b\b \nFinished compiling translation files. \n')


class CleanTranslations(Command):
    description = 'Cleans translations files.'
    user_options = [
        ('all', 'a', 'Remove all build output, not just temporary by-products')
    ]
    boolean_options = ['all']

    def initialize_options(self):
        self.all = None

    def finalize_options(self):
        self.set_undefined_options('clean', ('all', 'all'))

    def run(self):
        for path in (desktop_data, appdata_data):
            if os.path.isfile(path):
                print('Deleting %s' % path)
                os.remove(path)


class BuildPlugins(Command):
    description = 'Build plugins into .eggs'

    user_options = [
        ('install-dir=', None, 'develop install folder'),
        ('develop', 'D', 'Compile plugins in develop mode'),
    ]
    boolean_options = ['develop']

    def initialize_options(self):
        self.install_dir = None
        self.develop = False

    def finalize_options(self):
        pass

    def run(self):
        # Build the plugin eggs
        plugin_path = 'deluge/plugins/*'

        for path in glob.glob(plugin_path):
            if os.path.exists(os.path.join(path, 'setup.py')):
                if self.develop and self.install_dir:
                    os.system(
                        'cd '
                        + path
                        + '&& '
                        + sys.executable
                        + ' setup.py develop --install-dir=%s' % self.install_dir
                    )
                elif self.develop:
                    os.system(
                        'cd ' + path + '&& ' + sys.executable + ' setup.py develop'
                    )
                else:
                    os.system(
                        'cd '
                        + path
                        + '&& '
                        + sys.executable
                        + ' setup.py bdist_egg -d ..'
                    )


class CleanPlugins(Command):
    description = 'Cleans the plugin folders'
    user_options = [
        ('all', 'a', 'Remove all build output, not just temporary by-products')
    ]
    boolean_options = ['all']

    def initialize_options(self):
        self.all = None

    def finalize_options(self):
        self.set_undefined_options('clean', ('all', 'all'))

    def run(self):
        print('Cleaning the plugin\'s folders...')

        plugin_path = 'deluge/plugins/*'

        for path in glob.glob(plugin_path):
            if os.path.exists(os.path.join(path, 'setup.py')):
                c = 'cd ' + path + ' && ' + sys.executable + ' setup.py clean'
                if self.all:
                    c += ' -a'
                print('Calling \'%s\'' % c)
                os.system(c)

            # Delete the .eggs
            if path[-4:] == '.egg':
                print('Deleting egg file "%s"' % path)
                os.remove(path)

            # Delete the .egg-link
            if path[-9:] == '.egg-link':
                print('Deleting egg link "%s"' % path)
                os.remove(path)

        egg_info_dir_path = 'deluge/plugins/*/*.egg-info'

        for path in glob.glob(egg_info_dir_path):
            # Delete the .egg-info's directories
            if path[-9:] == '.egg-info':
                print('Deleting %s' % path)
                for fpath in os.listdir(path):
                    os.remove(os.path.join(path, fpath))
                os.removedirs(path)


class EggInfoPlugins(Command):
    description = 'Create .egg-info directories for plugins'

    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        # Build the plugin eggs
        plugin_path = 'deluge/plugins/*'

        for path in glob.glob(plugin_path):
            if os.path.exists(os.path.join(path, 'setup.py')):
                os.system('cd ' + path + '&& ' + sys.executable + ' setup.py egg_info')


class Build(_build):
    sub_commands = [
        ('build_webui', None),
        ('build_trans', None),
        ('build_plugins', None),
    ] + _build.sub_commands

    def run(self):
        # Run all sub-commands (at least those that need to be run).
        _build.run(self)
        try:
            from deluge._libtorrent import LT_VERSION

            print(f'Info: Found libtorrent ({LT_VERSION}) installed.')
        except ImportError as ex:
            print('Warning: libtorrent (libtorrent-rasterbar) not found: %s' % ex)


class InstallData(_install_data):
    """Custom class to fix `setup install` copying data files to incorrect location. (Bug #1389)"""

    def finalize_options(self):
        self.install_dir = None
        self.set_undefined_options(
            'install',
            ('install_data', 'install_dir'),
            ('root', 'root'),
            ('force', 'force'),
        )

    def run(self):
        _install_data.run(self)


class Clean(_clean):
    sub_commands = _clean.sub_commands + [
        ('clean_plugins', None),
        ('clean_trans', None),
        ('clean_webui', None),
    ]

    def run(self):
        # Remove deluge egg-info.
        root_egg_info_dir_path = 'deluge*.egg-info'
        for path in glob.glob(root_egg_info_dir_path):
            print('Deleting %s' % path)
            for fpath in os.listdir(path):
                os.remove(os.path.join(path, fpath))
            os.removedirs(path)

        # Run all sub-commands (at least those that need to be run)
        for cmd_name in self.get_sub_commands():
            self.run_command(cmd_name)
        _clean.run(self)


cmdclass = {
    'build': Build,
    'build_webui': BuildWebUI,
    'build_trans': BuildTranslations,
    'build_plugins': BuildPlugins,
    'build_docs': BuildDoc,
    'spellcheck_docs': BuildDoc,
    'install_data': InstallData,
    'clean_plugins': CleanPlugins,
    'clean_trans': CleanTranslations,
    'clean_docs': CleanDocs,
    'clean_webui': CleanWebUI,
    'clean': Clean,
    'egg_info_plugins': EggInfoPlugins,
    'test': PyTest,
}


if not windows_check() and not osx_check():
    for icon_path in glob.glob('deluge/ui/data/icons/hicolor/*x*'):
        size = os.path.basename(icon_path)
        icons = glob.glob(os.path.join(icon_path, 'apps', 'deluge*.png'))
        _data_files.append((f'share/icons/hicolor/{size}/apps', icons))
    _data_files.extend(
        [
            (
                'share/icons/hicolor/scalable/apps',
                ['deluge/ui/data/icons/hicolor/scalable/apps/deluge.svg'],
            ),
            ('share/pixmaps', ['deluge/ui/data/pixmaps/deluge.png']),
            (
                'share/man/man1',
                [
                    'docs/man/deluge.1',
                    'docs/man/deluged.1',
                    'docs/man/deluge-gtk.1',
                    'docs/man/deluge-web.1',
                    'docs/man/deluge-console.1',
                ],
            ),
        ]
    )
    if os.path.isfile(desktop_data):
        _data_files.append(('share/applications', [desktop_data]))
    if os.path.isfile(appdata_data):
        _data_files.append(('share/appdata', [appdata_data]))


# Entry Points
_entry_points['console_scripts'] = [
    'deluge-console = deluge.ui.console:start',
]

# On Windows use gui_scripts to hide cmd popup (no effect on Linux/MacOS)
_entry_points['gui_scripts'] = [
    'deluge = deluge.ui.ui_entry:start_ui',
    'deluge-gtk = deluge.ui.gtk3:start',
    'deluge-web = deluge.ui.web:start',
    'deluged = deluge.core.daemon_entry:start_daemon',
]

# Provide Windows 'debug' exes for stdin/stdout e.g. logging/errors
if windows_check():
    _entry_points['console_scripts'].extend(
        [
            'deluge-debug = deluge.ui.ui_entry:start_ui',
            'deluge-web-debug = deluge.ui.web:start',
            'deluged-debug = deluge.core.daemon_entry:start_daemon',
        ]
    )

_entry_points['deluge.ui'] = [
    'console = deluge.ui.console:Console',
    'web = deluge.ui.web:Web',
    'gtk = deluge.ui.gtk3:Gtk',
]


_package_data['deluge'] = [
    'ui/data/pixmaps/*.png',
    'ui/data/pixmaps/*.svg',
    'ui/data/pixmaps/*.ico',
    'ui/data/pixmaps/*.gif',
    'ui/data/pixmaps/flags/*.png',
    'plugins/*.egg',
    'i18n/*/LC_MESSAGES/*.mo',
]
_package_data['deluge.ui.web'] = [
    'index.html',
    'css/*.css',
    'icons/*.png',
    'images/*.gif',
    'images/*.png',
    'js/*.js',
    'js/extjs/*.js',
    'render/*.html',
    'themes/css/*.css',
    'themes/images/*/*.gif',
    'themes/images/*/*.png',
    'themes/images/*/*/*.gif',
    'themes/images/*/*/*.png',
]
_package_data['deluge.ui.gtk3'] = ['glade/*.ui']

setup_requires = ['setuptools', 'wheel']
install_requires = [
    'twisted[tls]>=17.1',
    # Add pyasn1 for setuptools workaround:
    #   https://github.com/pypa/setuptools/issues/1510
    'pyasn1',
    'rencode',
    'pyopenssl',
    'pyxdg',
    'mako',
    'setuptools',
    "pywin32; sys_platform == 'win32'",
    "certifi; sys_platform == 'win32'",
    'zope.interface',
]
extras_require = {
    'all': [
        'setproctitle',
        'pillow',
        'chardet',
        'ifaddr',
    ]
}

# Main setup
setup(
    name='deluge',
    version=_version,
    fullname='Deluge BitTorrent Client',
    description='BitTorrent Client',
    author='Deluge Team',
    maintainer='Calum Lind',
    maintainer_email='calumlind+deluge@gmail.com',
    keywords='torrent bittorrent p2p fileshare filesharing',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://deluge-torrent.org',
    project_urls={
        'GitHub (mirror)': 'https://github.com/deluge-torrent/deluge',
        'Sourcecode': 'http://git.deluge-torrent.org/deluge',
        'Issues': 'https://dev.deluge-torrent.org/report/1',
        'Discussion': 'https://forum.deluge-torrent.org',
        'Documentation': 'https://deluge.readthedocs.io',
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Environment :: Web Environment',
        'Environment :: X11 Applications :: GTK',
        'Framework :: Twisted',
        'Intended Audience :: End Users/Desktop',
        (
            'License :: OSI Approved :: '
            'GNU General Public License v3 or later (GPLv3+)'
        ),
        'Programming Language :: Python',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Topic :: Internet',
    ],
    python_requires='>=3.6',
    license='GPLv3+',
    cmdclass=cmdclass,
    setup_requires=setup_requires,
    install_requires=install_requires,
    extras_require=extras_require,
    data_files=_data_files,
    package_data=_package_data,
    exclude_package_data=_exclude_package_data,
    packages=find_packages(exclude=['deluge.plugins.*', 'deluge.tests']),
    entry_points=_entry_points,
)



#
# Copyright (C) 2008-2009 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging
import os
import sys
from base64 import b64encode
from glob import glob
from tempfile import mkstemp
from urllib.parse import urlparse
from urllib.request import url2pathname

import rencode
import twisted.internet.error
from twisted.internet import reactor
from twisted.internet.protocol import ClientFactory, Factory, Protocol, connectionDone

import deluge.component as component
from deluge.common import decode_bytes, is_magnet, is_url, windows_check
from deluge.configmanager import ConfigManager, get_config_dir
from deluge.ui.client import client

log = logging.getLogger(__name__)


class IPCProtocolServer(Protocol):
    def __init__(self):
        pass

    def dataReceived(self, data):  # NOQA: N802
        config = ConfigManager('gtk3ui.conf')
        data = rencode.loads(data, decode_utf8=True)
        if not data or config['focus_main_window_on_add']:
            component.get('MainWindow').present()
        process_args(data)


class IPCProtocolClient(Protocol):
    def __init__(self):
        pass

    def connectionMade(self):  # NOQA: N802
        self.transport.write(rencode.dumps(self.factory.args))
        self.transport.loseConnection()

    def connectionLost(self, reason=connectionDone):  # NOQA: N802
        reactor.stop()
        self.factory.stop = True


class IPCClientFactory(ClientFactory):
    protocol = IPCProtocolClient

    def __init__(self):
        self.stop = False

    def clientConnectionFailed(self, connector, reason):  # NOQA: N802
        log.warning('Connection to running instance failed.')
        reactor.stop()


class IPCInterface(component.Component):
    def __init__(self, args):
        component.Component.__init__(self, 'IPCInterface')
        self.listener = None
        ipc_dir = get_config_dir('ipc')
        if not os.path.exists(ipc_dir):
            os.makedirs(ipc_dir)
        socket = os.path.join(ipc_dir, 'deluge-gtk')
        if windows_check():
            # If we're on windows we need to check the global mutex to see if deluge is
            # already running.
            import win32api
            import win32event
            import winerror

            self.mutex = win32event.CreateMutex(None, False, 'deluge')
            if win32api.GetLastError() != winerror.ERROR_ALREADY_EXISTS:
                # Create listen socket
                self.factory = Factory()
                self.factory.protocol = IPCProtocolServer
                import random

                port = random.randrange(20000, 65535)
                self.listener = reactor.listenTCP(port, self.factory)
                # Store the port number in the socket file
                with open(socket, 'w') as _file:
                    _file.write(str(port))
                # We need to process any args when starting this process
                process_args(args)
            else:
                # Send to existing deluge process
                with open(socket) as _file:
                    port = int(_file.readline())
                self.factory = ClientFactory()
                self.factory.args = args
                self.factory.protocol = IPCProtocolClient
                reactor.connectTCP('127.0.0.1', port, self.factory)
                reactor.run()
                sys.exit(0)
        else:
            # Find and remove any restart tempfiles
            restart_tempfile = glob(os.path.join(ipc_dir, 'restart.*'))
            for f in restart_tempfile:
                os.remove(f)
            lockfile = socket + '.lock'
            log.debug('Checking if lockfile exists: %s', lockfile)
            if os.path.lexists(lockfile):

                def delete_lockfile():
                    log.debug('Delete stale lockfile.')
                    try:
                        os.remove(lockfile)
                        os.remove(socket)
                    except OSError as ex:
                        log.error('Failed to delete lockfile: %s', ex)

                try:
                    os.kill(int(os.readlink(lockfile)), 0)
                except OSError:
                    delete_lockfile()
                else:
                    if restart_tempfile:
                        log.warning(
                            'Found running PID but it is not a Deluge process, removing lockfile...'
                        )
                        delete_lockfile()
            try:
                self.factory = Factory()
                self.factory.protocol = IPCProtocolServer
                self.listener = reactor.listenUNIX(socket, self.factory, wantPID=True)
            except twisted.internet.error.CannotListenError as ex:
                log.info(
                    'Deluge is already running! Sending arguments to running instance...'
                )
                self.factory = IPCClientFactory()
                self.factory.args = args
                reactor.connectUNIX(socket, self.factory, checkPID=True)
                reactor.run()
                if self.factory.stop:
                    log.info('Success sending arguments to running Deluge.')
                    from gi.repository.Gdk import notify_startup_complete

                    notify_startup_complete()
                    sys.exit(0)
                else:
                    if restart_tempfile:
                        log.error('Deluge restart failed: %s', ex)
                        sys.exit(1)
                    else:
                        log.warning('Restarting Deluge... (%s)', ex)
                        # Create a tempfile to keep track of restart
                        mkstemp(prefix='restart.', dir=ipc_dir)
                        os.execv(sys.argv[0], sys.argv)
            else:
                process_args(args)

    def shutdown(self):
        if windows_check():
            import win32api

            win32api.CloseHandle(self.mutex)
        if self.listener:
            return self.listener.stopListening()


def process_args(args):
    """Process arguments sent to already running Deluge"""
    # Make sure args is a list
    args = list(args)
    log.debug('Processing args from other process: %s', args)
    if not client.connected():
        # We're not connected so add these to the queue
        log.debug('Not connected to host.. Adding to queue.')
        component.get('QueuedTorrents').add_to_queue(args)
        return
    config = ConfigManager('gtk3ui.conf')

    for arg in args:
        if not arg.strip():
            continue
        log.debug('arg: %s', arg)

        if is_url(arg):
            log.debug('Attempting to add url (%s) from external source...', arg)
            if config['interactive_add']:
                component.get('AddTorrentDialog').add_from_url(arg)
                component.get('AddTorrentDialog').show(config['focus_add_dialog'])
            else:
                client.core.add_torrent_url(arg, None)

        elif is_magnet(arg):
            log.debug('Attempting to add magnet (%s) from external source...', arg)
            if config['interactive_add']:
                component.get('AddTorrentDialog').add_from_magnets([arg])
                component.get('AddTorrentDialog').show(config['focus_add_dialog'])
            else:
                client.core.add_torrent_magnet(arg, {})

        else:
            log.debug('Attempting to add file (%s) from external source...', arg)
            if urlparse(arg).scheme == 'file':
                arg = url2pathname(urlparse(arg).path)
            path = os.path.abspath(decode_bytes(arg))

            if not os.path.exists(path):
                log.error('No such file: %s', path)
                continue

            if config['interactive_add']:
                component.get('AddTorrentDialog').add_from_files([path])
                component.get('AddTorrentDialog').show(config['focus_add_dialog'])
            else:
                with open(path, 'rb') as _file:
                    filedump = b64encode(_file.read())
                client.core.add_torrent_file(os.path.split(path)[-1], filedump, None)



#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import json
import logging
import os
from codecs import getwriter

import pytest
import pytest_twisted
from twisted.internet import task

from deluge.common import JSON_FORMAT
from deluge.config import Config
from deluge.ui.hostlist import mask_hosts_password

DEFAULTS = {
    'string': 'foobar',
    'int': 1,
    'float': 0.435,
    'bool': True,
    'unicode': 'foobar',
    'password': 'abc123*\\[!]?/<>#{@}=|"+$%(^)~',
    'hosts': [
        ('host1', 'port', '', 'password1234'),
        ('host2', 'port', '', 'password5678'),
    ],
}


LOGGER = logging.getLogger(__name__)


class TestConfig:
    def test_init(self):
        config = Config('test.conf', defaults=DEFAULTS, config_dir=self.config_dir)
        assert DEFAULTS == config.config

        config = Config('test.conf', config_dir=self.config_dir)
        assert {} == config.config

    def test_set_get_item(self):
        config = Config('test.conf', config_dir=self.config_dir)
        config['foo'] = 1
        assert config['foo'] == 1
        with pytest.raises(ValueError):
            config.set_item('foo', 'bar')

        config['foo'] = 2
        assert config.get_item('foo') == 2

        config['foo'] = '3'
        assert config.get_item('foo') == 3

        config['unicode'] = 'ВИДЕОФИЛЬМЫ'
        assert config['unicode'] == 'ВИДЕОФИЛЬМЫ'

        config['unicode'] = b'foostring'
        assert not isinstance(config.get_item('unicode'), bytes)

        config._save_timer.cancel()

    def test_set_get_item_none(self):
        config = Config('test.conf', config_dir=self.config_dir)

        config['foo'] = None
        assert config['foo'] is None
        assert isinstance(config['foo'], type(None))

        config['foo'] = 1
        assert config.get('foo') == 1

        config['foo'] = None
        assert config['foo'] is None

        config['bar'] = None
        assert config['bar'] is None

        config['bar'] = None
        assert config['bar'] is None

        config._save_timer.cancel()

    @pytest_twisted.ensureDeferred
    async def test_on_changed_callback(self, mock_callback):
        config = Config('test.conf', config_dir=self.config_dir)
        config.register_change_callback(mock_callback)
        config['foo'] = 1
        assert config['foo'] == 1
        await mock_callback.deferred
        mock_callback.assert_called_once_with('foo', 1)

    @pytest_twisted.ensureDeferred
    async def test_key_function_callback(self, mock_callback):
        config = Config(
            'test.conf', defaults={'foo': 1, 'bar': 1}, config_dir=self.config_dir
        )

        assert config['foo'] == 1
        config.register_set_function('foo', mock_callback)
        await mock_callback.deferred
        mock_callback.assert_called_once_with('foo', 1)

        mock_callback.reset_mock()
        config.register_set_function('bar', mock_callback, apply_now=False)
        mock_callback.assert_not_called()
        config['bar'] = 2
        await mock_callback.deferred
        mock_callback.assert_called_once_with('bar', 2)

    def test_get(self):
        config = Config('test.conf', config_dir=self.config_dir)
        config['foo'] = 1
        assert config.get('foo') == 1
        assert config.get('foobar') is None
        assert config.get('foobar', 2) == 2
        config['foobar'] = 5
        assert config.get('foobar', 2) == 5

    def test_set_log_mask_funcs(self, caplog):
        """Test mask func masks key in log"""
        caplog.set_level(logging.DEBUG)
        config = Config(
            'test.conf',
            config_dir=self.config_dir,
            log_mask_funcs={'hosts': mask_hosts_password},
        )
        config['hosts'] = DEFAULTS['hosts']
        assert isinstance(config['hosts'], list)
        assert 'host1' in caplog.text
        assert 'host2' in caplog.text
        assert 'password1234' not in caplog.text
        assert 'password5678' not in caplog.text
        assert '*' * 10 in caplog.text

    def test_load_log_mask_funcs(self, caplog):
        """Test mask func masks key in log"""
        with open(os.path.join(self.config_dir, 'test.conf'), 'wb') as _file:
            json.dump(DEFAULTS, getwriter('utf8')(_file), **JSON_FORMAT)

        config = Config(
            'test.conf',
            config_dir=self.config_dir,
            log_mask_funcs={'hosts': mask_hosts_password},
        )
        with caplog.at_level(logging.DEBUG):
            config.load(os.path.join(self.config_dir, 'test.conf'))
        assert 'host1' in caplog.text
        assert 'host2' in caplog.text
        assert 'foobar' in caplog.text
        assert 'password1234' not in caplog.text
        assert 'password5678' not in caplog.text
        assert '*' * 10 in caplog.text

    def test_load(self):
        def check_config():
            config = Config('test.conf', config_dir=self.config_dir)

            assert config['string'] == 'foobar'
            assert config['float'] == 0.435
            assert config['password'] == 'abc123*\\[!]?/<>#{@}=|"+$%(^)~'

        # Test opening a previous 1.2 config file of just a json object
        with open(os.path.join(self.config_dir, 'test.conf'), 'wb') as _file:
            json.dump(DEFAULTS, getwriter('utf8')(_file), **JSON_FORMAT)

        check_config()

        # Test opening a previous 1.2 config file of having the format versions
        # as ints
        with open(os.path.join(self.config_dir, 'test.conf'), 'wb') as _file:
            _file.write(b'1\n')
            _file.write(b'1\n')
            json.dump(DEFAULTS, getwriter('utf8')(_file), **JSON_FORMAT)

        check_config()

        # Test the 1.2 config format
        version = {'format': 1, 'file': 1}
        with open(os.path.join(self.config_dir, 'test.conf'), 'wb') as _file:
            json.dump(version, getwriter('utf8')(_file), **JSON_FORMAT)
            json.dump(DEFAULTS, getwriter('utf8')(_file), **JSON_FORMAT)

        check_config()

    def test_save(self):
        config = Config('test.conf', defaults=DEFAULTS, config_dir=self.config_dir)
        # We do this twice because the first time we need to save the file to disk
        # and the second time we do a compare and we should not write
        ret = config.save()
        assert ret
        ret = config.save()
        assert ret

        config['string'] = 'baz'
        config['int'] = 2
        ret = config.save()
        assert ret
        del config

        config = Config('test.conf', defaults=DEFAULTS, config_dir=self.config_dir)
        assert config['string'] == 'baz'
        assert config['int'] == 2

    def test_save_timer(self):
        clock = task.Clock()

        config = Config('test.conf', defaults=DEFAULTS, config_dir=self.config_dir)
        config.callLater = clock.callLater
        config['string'] = 'baz'
        config['int'] = 2
        assert config._save_timer.active()

        # Timeout set for 5 seconds in config, so lets move clock by 5 seconds
        clock.advance(5)

        def check_config(config):
            assert not config._save_timer.active()
            del config
            config = Config('test.conf', defaults=DEFAULTS, config_dir=self.config_dir)
            assert config['string'] == 'baz'
            assert config['int'] == 2

        check_config(config)

    def test_find_json_objects(self):
        s = """{
  "file": 1,
  "format": 1
}{
  "ssl": true,
  "enabled": false,
  "port": 8115
}\n"""

        from deluge.config import find_json_objects

        objects = find_json_objects(s)
        assert len(objects) == 2

    def test_find_json_objects_curly_brace(self):
        """Test with string containing curly brace"""
        s = """{
  "file": 1,
  "format": 1
}{
  "ssl": true,
  "enabled": false,
  "port": 8115,
  "password": "abc{def"
}"""

        from deluge.config import find_json_objects

        objects = find_json_objects(s)
        assert len(objects) == 2

    def test_find_json_objects_double_quote(self):
        """Test with string containing double quote"""
        s = r"""{
  "file": 1,
  "format": 1
}{
  "ssl": true,
  "enabled": false,
  "port": 8115,
  "password": "abc\"def"
}
"""

        from deluge.config import find_json_objects

        objects = find_json_objects(s)
        assert len(objects) == 2



#
# Copyright (C) 2007-2008 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging

from gi.repository import Gtk
from gi.repository.GLib import timeout_add

import deluge.component as component
from deluge.common import fsize, fspeed, get_pixmap
from deluge.configmanager import ConfigManager
from deluge.ui.client import client

from .common import build_menu_radio_list
from .dialogs import OtherDialog

log = logging.getLogger(__name__)


class StatusBarItem:
    def __init__(
        self,
        image=None,
        stock=None,
        icon=None,
        text=None,
        markup=False,
        callback=None,
        tooltip=None,
    ):
        self._widgets = []
        self._ebox = Gtk.EventBox()
        self._hbox = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, spacing=3)
        self._image = Gtk.Image()
        self._label = Gtk.Label()
        if image or icon or stock:
            self._hbox.add(self._image)
        self._hbox.add(self._label)
        self._ebox.add(self._hbox)

        # Add image from file or stock
        if image:
            self.set_image_from_file(image)
        if stock:
            self.set_image_from_stock(stock)
        if icon:
            self.set_image_from_icon(icon)

        # Add text
        if markup:
            self.set_markup(text)
        else:
            self.set_text(text)

        if callback is not None:
            self.set_callback(callback)

        if tooltip:
            self.set_tooltip(tooltip)

        self.show_all()

    def set_callback(self, callback):
        self._ebox.connect('button-press-event', callback)

    def show_all(self):
        self._ebox.show()
        self._hbox.show()
        self._image.show()

    def set_image_from_file(self, image):
        self._image.set_from_file(image)

    def set_image_from_stock(self, stock):
        self._image.set_from_stock(stock, Gtk.IconSize.MENU)

    def set_image_from_icon(self, icon):
        self._image.set_from_icon_name(icon, Gtk.IconSize.MENU)

    def set_text(self, text):
        if not text:
            self._label.hide()
        elif self._label.get_text() != text:
            self._label.set_text(text)
            self._label.show()

    def set_markup(self, text):
        if not text:
            self._label.hide()
        elif self._label.get_text() != text:
            self._label.set_markup(text)
            self._label.show()

    def set_tooltip(self, tip):
        if self._ebox.get_tooltip_text() != tip:
            self._ebox.set_tooltip_text(tip)

    def get_widgets(self):
        return self._widgets

    def get_eventbox(self):
        return self._ebox

    def get_text(self):
        return self._label.get_text()


class StatusBar(component.Component):
    def __init__(self):
        component.Component.__init__(self, 'StatusBar', interval=3)
        main_builder = component.get('MainWindow').get_builder()
        self.statusbar = main_builder.get_object('statusbar')
        self.config = ConfigManager('gtk3ui.conf')

        # Status variables that are updated via callback
        self.max_connections_global = -1
        self.num_connections = 0
        self.max_download_speed = -1.0
        self.download_rate = ''
        self.max_upload_speed = -1.0
        self.upload_rate = ''
        self.dht_nodes = 0
        self.dht_status = False
        self.health = False
        self.download_protocol_rate = 0.0
        self.upload_protocol_rate = 0.0

        self.config_value_changed_dict = {
            'max_connections_global': self._on_max_connections_global,
            'max_download_speed': self._on_max_download_speed,
            'max_upload_speed': self._on_max_upload_speed,
            'dht': self._on_dht,
        }
        self.current_warnings = []
        # Add hbox to the statusbar after removing the initial label widget
        self.hbox = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, spacing=10)
        self.hbox.set_margin_top(2)
        self.hbox.set_margin_bottom(3)
        frame = self.statusbar.get_children()[0]
        frame.remove(frame.get_children()[0])
        frame.add(self.hbox)
        self.statusbar.show_all()
        # Create the not connected item
        self.not_connected_item = StatusBarItem(
            icon='network-offline-symbolic',
            text=_('Not Connected'),
            callback=self._on_notconnected_item_clicked,
        )
        # Show the not connected status bar
        self.show_not_connected()

        # Hide if necessary
        self.visible(self.config['show_statusbar'])

        client.register_event_handler(
            'ConfigValueChangedEvent', self.on_configvaluechanged_event
        )

    def start(self):
        # Add in images and labels
        self.remove_item(self.not_connected_item)

        self.connections_item = self.add_item(
            icon='network-transmit-receive-symbolic',
            callback=self._on_connection_item_clicked,
            tooltip=_('Connections (Limit)'),
            pack_start=True,
        )

        self.download_item = self.add_item(
            image=get_pixmap('downloading16.png'),
            callback=self._on_download_item_clicked,
            tooltip=_('Download Speed (Limit)'),
            pack_start=True,
        )

        self.upload_item = self.add_item(
            image=get_pixmap('seeding16.png'),
            callback=self._on_upload_item_clicked,
            tooltip=_('Upload Speed (Limit)'),
            pack_start=True,
        )

        self.traffic_item = self.add_item(
            image=get_pixmap('traffic16.png'),
            callback=self._on_traffic_item_clicked,
            tooltip=_('Protocol Traffic (Down:Up)'),
            pack_start=True,
        )

        self.dht_item = StatusBarItem(
            image=get_pixmap('dht16.png'), tooltip=_('DHT Nodes')
        )

        self.diskspace_item = self.add_item(
            icon='drive-harddisk-symbolic',
            callback=self._on_diskspace_item_clicked,
            tooltip=_('Free Disk Space'),
            pack_start=True,
        )

        self.external_ip_item = self.add_item(
            tooltip=_('External IP Address'),
            text=_('<b>IP</b> <small>%s</small>') % _('n/a'),
            markup=True,
            pack_start=True,
        )

        self.health_item = self.add_item(
            icon='network-error-symbolic',
            text=_('<b><small>Port Issue</small></b>'),
            markup=True,
            tooltip=_('No incoming connections, check port forwarding'),
            callback=self._on_health_icon_clicked,
        )

        self.health = False

        def update_config_values(configs):
            self._on_max_connections_global(configs['max_connections_global'])
            self._on_max_download_speed(configs['max_download_speed'])
            self._on_max_upload_speed(configs['max_upload_speed'])
            self._on_dht(configs['dht'])

        # Get some config values
        client.core.get_config_values(
            ['max_connections_global', 'max_download_speed', 'max_upload_speed', 'dht']
        ).addCallback(update_config_values)

    def stop(self):
        # When stopped, we just show the not connected thingy
        try:
            self.remove_item(self.connections_item)
            self.remove_item(self.dht_item)
            self.remove_item(self.download_item)
            self.remove_item(self.upload_item)
            self.remove_item(self.not_connected_item)
            self.remove_item(self.health_item)
            self.remove_item(self.traffic_item)
            self.remove_item(self.diskspace_item)
            self.remove_item(self.external_ip_item)
        except Exception as ex:
            log.debug('Unable to remove StatusBar item: %s', ex)
        self.show_not_connected()

    def visible(self, visible):
        if visible:
            self.statusbar.show()
        else:
            self.statusbar.hide()

        self.config['show_statusbar'] = visible

    def show_not_connected(self):
        self.hbox.pack_start(self.not_connected_item.get_eventbox(), False, False, 0)

    def add_item(
        self,
        image=None,
        stock=None,
        icon=None,
        text=None,
        markup=False,
        callback=None,
        tooltip=None,
        pack_start=False,
    ):
        """Adds an item to the status bar"""
        # The return tuple.. we return whatever widgets we add
        item = StatusBarItem(image, stock, icon, text, markup, callback, tooltip)
        if pack_start:
            self.hbox.pack_start(item.get_eventbox(), False, False, 0)
        else:
            self.hbox.pack_end(item.get_eventbox(), False, False, 0)
        return item

    def remove_item(self, item):
        """Removes an item from the statusbar"""
        if item.get_eventbox() in self.hbox.get_children():
            try:
                self.hbox.remove(item.get_eventbox())
            except Exception as ex:
                log.debug('Unable to remove widget: %s', ex)

    def add_timeout_item(
        self, seconds=3, image=None, stock=None, icon=None, text=None, callback=None
    ):
        """Adds an item to the StatusBar for seconds"""
        item = self.add_item(image, stock, icon, text, callback)
        # Start a timer to remove this item in seconds
        timeout_add(seconds * 1000, self.remove_item, item)

    def display_warning(self, text, callback=None):
        """Displays a warning to the user in the status bar"""
        if text not in self.current_warnings:
            item = self.add_item(
                icon='dialog-warning-symbolic', text=text, callback=callback
            )
            self.current_warnings.append(text)
            timeout_add(3000, self.remove_warning, item)

    def remove_warning(self, item):
        self.current_warnings.remove(item.get_text())
        self.remove_item(item)

    def clear_statusbar(self):
        def remove(child):
            self.hbox.remove(child)

        self.hbox.foreach(remove)

    def send_status_request(self):
        # Sends an async request for data from the core
        keys = [
            'peer.num_peers_connected',
            'upload_rate',
            'download_rate',
            'payload_upload_rate',
            'payload_download_rate',
            'net.sent_bytes',
            'net.recv_bytes',
            'net.sent_payload_bytes',
            'net.recv_payload_bytes',
        ]

        if self.dht_status:
            keys.append('dht.dht_nodes')

        if not self.health:
            keys.append('net.has_incoming_connections')

        client.core.get_session_status(keys).addCallback(self._on_get_session_status)
        client.core.get_free_space().addCallback(self._on_get_free_space)
        client.core.get_external_ip().addCallback(self._on_get_external_ip)

    def on_configvaluechanged_event(self, key, value):
        """
        This is called when we receive a ConfigValueChangedEvent from
        the core.
        """
        if key in self.config_value_changed_dict:
            self.config_value_changed_dict[key](value)

    def _on_max_connections_global(self, max_connections):
        self.max_connections_global = max_connections
        self.update_connections_label()

    def _on_dht(self, value):
        self.dht_status = value
        if value:
            self.hbox.pack_start(self.dht_item.get_eventbox(), False, False, 0)
            self.send_status_request()
        else:
            self.remove_item(self.dht_item)

    def _on_get_session_status(self, status):
        self.download_rate = fspeed(
            status['payload_download_rate'], precision=0, shortform=True
        )
        self.upload_rate = fspeed(
            status['payload_upload_rate'], precision=0, shortform=True
        )
        self.download_protocol_rate = (
            status['download_rate'] - status['payload_download_rate']
        ) // 1024
        self.upload_protocol_rate = (
            status['upload_rate'] - status['payload_upload_rate']
        ) // 1024
        self.num_connections = status['peer.num_peers_connected']
        self.update_download_label()
        self.update_upload_label()
        self.update_traffic_label()
        self.update_connections_label()

        if 'dht.dht_nodes' in status:
            self.dht_nodes = status['dht.dht_nodes']
            self.update_dht_label()

        if 'net.has_incoming_connections' in status:
            self.health = status['net.has_incoming_connections']
            if self.health:
                self.remove_item(self.health_item)

    def _on_get_free_space(self, space):
        if space >= 0:
            self.diskspace_item.set_markup(
                '<small>%s</small>' % fsize(space, shortform=True)
            )
        else:
            self.diskspace_item.set_markup(
                '<span foreground="red">' + _('Error') + '</span>'
            )

    def _on_max_download_speed(self, max_download_speed):
        self.max_download_speed = max_download_speed
        self.update_download_label()

    def _on_max_upload_speed(self, max_upload_speed):
        self.max_upload_speed = max_upload_speed
        self.update_upload_label()

    def _on_get_external_ip(self, external_ip):
        ip = external_ip if external_ip else _('n/a')
        self.external_ip_item.set_markup(_('<b>IP</b> <small>%s</small>') % ip)

    def update_connections_label(self):
        # Set the max connections label
        if self.max_connections_global < 0:
            label_string = '%s' % self.num_connections
        else:
            label_string = '{} <small>({})</small>'.format(
                self.num_connections,
                self.max_connections_global,
            )

        self.connections_item.set_markup(label_string)

        if self.num_connections:
            self.connections_item.set_image_from_icon(
                'network-transmit-receive-symbolic'
            )
        else:
            self.connections_item.set_image_from_icon('network-idle-symbolic')

    def update_dht_label(self):
        # Set the max connections label
        self.dht_item.set_markup('<small>%s</small>' % (self.dht_nodes))

    def update_download_label(self):
        # Set the download speed label
        if self.max_download_speed <= 0:
            label_string = self.download_rate
        else:
            label_string = '%s <small>(%i %s)</small>' % (
                self.download_rate,
                self.max_download_speed,
                _('K/s'),
            )

        self.download_item.set_markup(label_string)

    def update_upload_label(self):
        # Set the upload speed label
        if self.max_upload_speed <= 0:
            label_string = self.upload_rate
        else:
            label_string = '%s <small>(%i %s)</small>' % (
                self.upload_rate,
                self.max_upload_speed,
                _('K/s'),
            )

        self.upload_item.set_markup(label_string)

    def update_traffic_label(self):
        label_string = '<small>%i:%i %s</small>' % (
            self.download_protocol_rate,
            self.upload_protocol_rate,
            _('K/s'),
        )
        self.traffic_item.set_markup(label_string)

    def update(self):
        self.send_status_request()

    def set_limit_value(self, widget, core_key):
        log.debug('_on_set_unlimit_other %s', core_key)
        other_dialog_info = {
            'max_download_speed': (
                _('Download Speed Limit'),
                _('Set the maximum download speed'),
                _('K/s'),
                'downloading.svg',
                self.max_download_speed,
            ),
            'max_upload_speed': (
                _('Upload Speed Limit'),
                _('Set the maximum upload speed'),
                _('K/s'),
                'seeding.svg',
                self.max_upload_speed,
            ),
            'max_connections_global': (
                _('Incoming Connections'),
                _('Set the maximum incoming connections'),
                '',
                'network-transmit-receive-symbolic',
                self.max_connections_global,
            ),
        }

        def set_value(value):
            log.debug('value: %s', value)
            if value is None:
                return
            elif value == 0:
                value = -1
            # Set the config in the core
            if value != getattr(self, core_key):
                client.core.set_config({core_key: value})

        if widget.get_name() == 'unlimited':
            set_value(-1)
        elif widget.get_name() == 'other':

            def dialog_finished(response_id):
                if response_id == Gtk.ResponseType.OK:
                    set_value(dialog.get_value())

            dialog = OtherDialog(*other_dialog_info[core_key])
            dialog.run().addCallback(set_value)
        else:
            value = widget.get_children()[0].get_text().split(' ')[0]
            set_value(value)

    def _on_download_item_clicked(self, widget, event):
        self.menu = build_menu_radio_list(
            self.config['tray_download_speed_list'],
            self._on_set_download_speed,
            self.max_download_speed,
            _('K/s'),
            show_notset=True,
            show_other=True,
        )
        self.menu.show_all()
        self.menu.popup(None, None, None, None, event.button, event.time)

    def _on_set_download_speed(self, widget):
        log.debug('_on_set_download_speed')
        self.set_limit_value(widget, 'max_download_speed')

    def _on_upload_item_clicked(self, widget, event):
        self.menu = build_menu_radio_list(
            self.config['tray_upload_speed_list'],
            self._on_set_upload_speed,
            self.max_upload_speed,
            _('K/s'),
            show_notset=True,
            show_other=True,
        )
        self.menu.show_all()
        self.menu.popup(None, None, None, None, event.button, event.time)

    def _on_set_upload_speed(self, widget):
        log.debug('_on_set_upload_speed')
        self.set_limit_value(widget, 'max_upload_speed')

    def _on_connection_item_clicked(self, widget, event):
        self.menu = build_menu_radio_list(
            self.config['connection_limit_list'],
            self._on_set_connection_limit,
            self.max_connections_global,
            show_notset=True,
            show_other=True,
        )
        self.menu.show_all()
        self.menu.popup(None, None, None, None, event.button, event.time)

    def _on_set_connection_limit(self, widget):
        log.debug('_on_set_connection_limit')
        self.set_limit_value(widget, 'max_connections_global')

    def _on_health_icon_clicked(self, widget, event):
        component.get('Preferences').show('network')

    def _on_notconnected_item_clicked(self, widget, event):
        component.get('ConnectionManager').show()

    def _on_traffic_item_clicked(self, widget, event):
        component.get('Preferences').show('network')

    def _on_diskspace_item_clicked(self, widget, event):
        component.get('Preferences').show('downloads')



#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import os
import sys
import tarfile
from urllib.parse import quote_plus

import pytest

from deluge.common import (
    VersionSplit,
    archive_files,
    fdate,
    fpcnt,
    fpeer,
    fsize,
    fspeed,
    ftime,
    get_magnet_info,
    get_path_size,
    is_infohash,
    is_interface,
    is_interface_name,
    is_ip,
    is_ipv4,
    is_ipv6,
    is_magnet,
    is_url,
    windows_check,
)

from .common import get_test_data_file


class TestCommon:
    def test_fsize(self):
        assert fsize(0) == '0 B'
        assert fsize(100) == '100 B'
        assert fsize(1023) == '1023 B'
        assert fsize(1024) == '1.0 KiB'
        assert fsize(1048575) == '1024.0 KiB'
        assert fsize(1048576) == '1.0 MiB'
        assert fsize(1073741823) == '1024.0 MiB'
        assert fsize(1073741824) == '1.0 GiB'
        assert fsize(112245) == '109.6 KiB'
        assert fsize(110723441824) == '103.1 GiB'
        assert fsize(1099511627775) == '1024.0 GiB'
        assert fsize(1099511627777) == '1.0 TiB'
        assert fsize(766148267453245) == '696.8 TiB'

    def test_fpcnt(self):
        assert fpcnt(0.9311) == '93.11%'

    def test_fspeed(self):
        assert fspeed(43134) == '42.1 KiB/s'

    def test_fpeer(self):
        assert fpeer(10, 20) == '10 (20)'
        assert fpeer(10, -1) == '10'

    def test_ftime(self):
        assert ftime(0) == ''
        assert ftime(5) == '5s'
        assert ftime(100) == '1m 40s'
        assert ftime(3789) == '1h 3m'
        assert ftime(23011) == '6h 23m'
        assert ftime(391187) == '4d 12h'
        assert ftime(604800) == '1w 0d'
        assert ftime(13893086) == '22w 6d'
        assert ftime(59740269) == '1y 46w'
        assert ftime(61.25) == '1m 1s'
        assert ftime(119.9) == '1m 59s'

    def test_fdate(self):
        assert fdate(-1) == ''

    def test_is_url(self):
        assert is_url('http://deluge-torrent.org')
        assert not is_url('file://test.torrent')

    def test_is_magnet(self):
        assert is_magnet('magnet:?xt=urn:btih:SU5225URMTUEQLDXQWRB2EQWN6KLTYKN')
        assert not is_magnet(None)

    def test_is_infohash(self):
        assert is_infohash('2dc5d0e71a66fe69649a640d39cb00a259704973')

    def test_get_path_size(self):
        if windows_check() and sys.version_info < (3, 8):
            # https://bugs.python.org/issue1311
            pytest.skip('os.devnull returns False on Windows')
        assert get_path_size(os.devnull) == 0
        assert get_path_size('non-existant.file') == -1

    def test_is_ip(self):
        assert is_ip('192.0.2.0')
        assert not is_ip('192..0.0')
        assert is_ip('2001:db8::')
        assert not is_ip('2001:db8:')

    def test_is_ipv4(self):
        assert is_ipv4('192.0.2.0')
        assert not is_ipv4('192..0.0')

    def test_is_ipv6(self):
        assert is_ipv6('2001:db8::')
        assert not is_ipv6('2001:db8:')

    def test_is_interface_name(self):
        if windows_check():
            assert not is_interface_name('2001:db8:')
            assert not is_interface_name('{THIS0000-IS00-ONLY-FOR0-TESTING00000}')
        else:
            assert is_interface_name('lo')
            assert not is_interface_name('127.0.0.1')
            assert not is_interface_name('eth01101')

    def test_is_interface(self):
        if windows_check():
            assert is_interface('127.0.0.1')
            assert not is_interface('127')
            assert not is_interface('{THIS0000-IS00-ONLY-FOR0-TESTING00000}')
        else:
            assert is_interface('lo')
            assert is_interface('127.0.0.1')
            assert not is_interface('127.')
            assert not is_interface('eth01101')

    def test_version_split(self):
        assert VersionSplit('1.2.2') == VersionSplit('1.2.2')
        assert VersionSplit('1.2.1') < VersionSplit('1.2.2')
        assert VersionSplit('1.1.9') < VersionSplit('1.2.2')
        assert VersionSplit('1.2.2') > VersionSplit('1.2.1')
        assert VersionSplit('1.2.2') > VersionSplit('1.2.2-dev0')
        assert VersionSplit('1.2.2-dev') < VersionSplit('1.3.0-rc2')
        assert VersionSplit('1.2.2') > VersionSplit('1.2.2-rc2')
        assert VersionSplit('1.2.2-rc2-dev') < VersionSplit('1.2.2-rc2')
        assert VersionSplit('1.2.2-rc3') > VersionSplit('1.2.2-rc2')
        assert VersionSplit('0.14.9') == VersionSplit('0.14.9')
        assert VersionSplit('0.14.9') > VersionSplit('0.14.5')
        assert VersionSplit('0.14.10') >= VersionSplit('0.14.9')
        assert VersionSplit('1.4.0') > VersionSplit('1.3.900.dev123')
        assert VersionSplit('1.3.2rc2.dev1') < VersionSplit('1.3.2-rc2')
        assert VersionSplit('1.3.900.dev888') > VersionSplit('1.3.900.dev123')
        assert VersionSplit('1.4.0') > VersionSplit('1.4.0.dev123')
        assert VersionSplit('1.4.0.dev1') < VersionSplit('1.4.0')
        assert VersionSplit('1.4.0a1') < VersionSplit('1.4.0')

    def test_parse_human_size(self):
        from deluge.common import parse_human_size

        sizes = [
            ('1', 1),
            ('10 bytes', 10),
            ('2048 bytes', 2048),
            ('1MiB', 2 ** (10 * 2)),
            ('1 MiB', 2 ** (10 * 2)),
            ('1 GiB', 2 ** (10 * 3)),
            ('1 GiB', 2 ** (10 * 3)),
            ('1M', 10**6),
            ('1MB', 10**6),
            ('1 GB', 10**9),
            ('1 TB', 10**12),
        ]

        for human_size, byte_size in sizes:
            parsed = parse_human_size(human_size)
            assert parsed == byte_size, 'Mismatch when converting: %s' % human_size

    def test_archive_files(self):
        arc_filelist = [
            get_test_data_file('test.torrent'),
            get_test_data_file('deluge.png'),
        ]
        arc_filepath = archive_files('test-arc', arc_filelist)

        with tarfile.open(arc_filepath, 'r') as tar:
            for tar_info in tar:
                assert tar_info.isfile()
                assert tar_info.name in [
                    os.path.basename(arcf) for arcf in arc_filelist
                ]

    def test_archive_files_missing(self):
        """Archive exists even with file not found."""
        filelist = ['test.torrent', 'deluge.png', 'missing.file']
        arc_filepath = archive_files(
            'test-arc', [get_test_data_file(f) for f in filelist]
        )
        filelist.remove('missing.file')

        with tarfile.open(arc_filepath, 'r') as tar:
            assert tar.getnames() == filelist
            assert all(tarinfo.isfile() for tarinfo in tar)

    def test_archive_files_message(self):
        filelist = ['test.torrent', 'deluge.png']
        arc_filepath = archive_files(
            'test-arc', [get_test_data_file(f) for f in filelist], message='test'
        )

        result_files = filelist + ['archive_message.txt']
        with tarfile.open(arc_filepath, 'r') as tar:
            assert tar.getnames() == result_files
            for tar_info in tar:
                assert tar_info.isfile()
                if tar_info.name == 'archive_message.txt':
                    result = tar.extractfile(tar_info).read().decode()
                    assert result == 'test'

    def test_get_magnet_info_tiers(self):
        tracker1 = 'udp://tracker1.example.com'
        tracker2 = 'udp://tracker2.example.com'
        magnet = (
            'magnet:?xt=urn:btih:SU5225URMTUEQLDXQWRB2EQWN6KLTYKN'
            f'&tr.1={quote_plus(tracker1)}'
            f'&tr.2={quote_plus(tracker2)}'
        )
        result = get_magnet_info(magnet)
        assert result['info_hash'] == '953bad769164e8482c7785a21d12166f94b9e14d'
        assert result['trackers'][tracker1] == 1
        assert result['trackers'][tracker2] == 2



#
# Copyright (C) 2009 Andrew Resch <andrewresch@gmail.com>
#
# Basic plugin template created by:
# Copyright (C) 2008 Martijn Voncken <mvoncken@gmail.com>
# Copyright (C) 2007-2009 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging
import time

from twisted.internet import reactor

import deluge.component as component
import deluge.configmanager
from deluge.core.rpcserver import export
from deluge.event import DelugeEvent
from deluge.plugins.pluginbase import CorePluginBase

log = logging.getLogger(__name__)

DEFAULT_PREFS = {
    'low_down': -1.0,
    'low_up': -1.0,
    'low_active': -1,
    'low_active_down': -1,
    'low_active_up': -1,
    'button_state': [[0] * 7 for dummy in range(24)],
}

STATES = {0: 'Green', 1: 'Yellow', 2: 'Red'}

CONTROLLED_SETTINGS = [
    'max_download_speed',
    'max_upload_speed',
    'max_active_limit',
    'max_active_downloading',
    'max_active_seeding',
]


class SchedulerEvent(DelugeEvent):
    """
    Emitted when a schedule state changes.
    """

    def __init__(self, colour):
        """
        :param colour: str, the current scheduler state
        """
        self._args = [colour]


class Core(CorePluginBase):
    def enable(self):
        # Create the defaults with the core config
        core_config = component.get('Core').config
        DEFAULT_PREFS['low_down'] = core_config['max_download_speed']
        DEFAULT_PREFS['low_up'] = core_config['max_upload_speed']
        DEFAULT_PREFS['low_active'] = core_config['max_active_limit']
        DEFAULT_PREFS['low_active_down'] = core_config['max_active_downloading']
        DEFAULT_PREFS['low_active_up'] = core_config['max_active_seeding']

        self.config = deluge.configmanager.ConfigManager(
            'scheduler.conf', DEFAULT_PREFS
        )

        self.state = self.get_state()

        # Apply the scheduling rules
        self.do_schedule(False)

        # Schedule the next do_schedule() call for on the next hour
        now = time.localtime(time.time())
        secs_to_next_hour = ((60 - now[4]) * 60) + (60 - now[5])
        self.timer = reactor.callLater(secs_to_next_hour, self.do_schedule)

        # Register for config changes so state isn't overridden
        component.get('EventManager').register_event_handler(
            'ConfigValueChangedEvent', self.on_config_value_changed
        )

    def disable(self):
        if self.timer.active():
            self.timer.cancel()
        component.get('EventManager').deregister_event_handler(
            'ConfigValueChangedEvent', self.on_config_value_changed
        )
        self.__apply_set_functions()

    def update(self):
        pass

    def on_config_value_changed(self, key, value):
        if key in CONTROLLED_SETTINGS:
            self.do_schedule(False)

    def __apply_set_functions(self):
        """
        Have the core apply it's bandwidth settings as specified in core.conf.
        """
        core_config = deluge.configmanager.ConfigManager('core.conf')
        for setting in CONTROLLED_SETTINGS:
            component.get('PreferencesManager').do_config_set_func(
                setting, core_config[setting]
            )
        # Resume the session if necessary
        component.get('Core').resume_session()

    def do_schedule(self, timer=True):
        """
        This is where we apply schedule rules.
        """

        state = self.get_state()

        if state == 'Green':
            # This is Green (Normal) so we just make sure we've applied the
            # global defaults
            self.__apply_set_functions()
        elif state == 'Yellow':
            # This is Yellow (Slow), so use the settings provided from the user
            settings = {
                'active_limit': self.config['low_active'],
                'active_downloads': self.config['low_active_down'],
                'active_seeds': self.config['low_active_up'],
                'download_rate_limit': int(self.config['low_down'] * 1024),
                'upload_rate_limit': int(self.config['low_up'] * 1024),
            }
            component.get('Core').apply_session_settings(settings)
            # Resume the session if necessary
            component.get('Core').resume_session()
        elif state == 'Red':
            # This is Red (Stop), so pause the libtorrent session
            component.get('Core').pause_session()

        if state != self.state:
            # The state has changed since last update so we need to emit an event
            self.state = state
            component.get('EventManager').emit(SchedulerEvent(self.state))

        if timer:
            # Call this again in 1 hour
            self.timer = reactor.callLater(3600, self.do_schedule)

    @export()
    def set_config(self, config):
        """Sets the config dictionary."""
        for key in config:
            self.config[key] = config[key]
        self.config.save()
        self.do_schedule(False)

    @export()
    def get_config(self):
        """Returns the config dictionary."""
        return self.config.config

    @export()
    def get_state(self):
        now = time.localtime(time.time())
        level = self.config['button_state'][now[3]][now[6]]
        return STATES[level]



#
# Copyright (C) 2011 Nick Lanham <nick@afternight.org>
# Copyright (C) 2008-2009 Ido Abramovich <ido.deluge@gmail.com>
# Copyright (C) 2009 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging
import os

from deluge.decorators import overrides
from deluge.ui.console.modes.basemode import InputKeyHandler
from deluge.ui.console.utils import colors
from deluge.ui.console.utils import curses_util as util
from deluge.ui.console.utils.format_utils import (
    delete_alt_backspace,
    remove_formatting,
    wrap_string,
)

try:
    import curses
except ImportError:
    pass

log = logging.getLogger(__name__)


class BaseField(InputKeyHandler):
    def __init__(self, parent=None, name=None, selectable=True, **kwargs):
        super().__init__()
        self.name = name
        self.parent = parent
        self.fmt_keys = {}
        self.set_fmt_key('font', 'ignore', kwargs)
        self.set_fmt_key('color', 'white,black', kwargs)
        self.set_fmt_key('color_end', 'white,black', kwargs)
        self.set_fmt_key('color_active', 'black,white', kwargs)
        self.set_fmt_key('color_unfocused', 'color', kwargs)
        self.set_fmt_key('color_unfocused_active', 'black,whitegrey', kwargs)
        self.set_fmt_key('font_active', 'font', kwargs)
        self.set_fmt_key('font_unfocused', 'font', kwargs)
        self.set_fmt_key('font_unfocused_active', 'font_active', kwargs)
        self.default_col = kwargs.get('col', -1)
        self._selectable = selectable
        self.value = None

    def selectable(self):
        return self.has_input() and not self.depend_skip() and self._selectable

    def set_fmt_key(self, key, default, kwargsdict=None):
        value = self.fmt_keys.get(default, default)
        if kwargsdict:
            value = kwargsdict.get(key, value)
        self.fmt_keys[key] = value

    def get_fmt_keys(self, focused, active, **kwargs):
        color_key = kwargs.get('color_key', 'color')
        font_key = 'font'
        if not focused:
            color_key += '_unfocused'
            font_key += '_unfocused'
        if active:
            color_key += '_active'
            font_key += '_active'
        return color_key, font_key

    def build_fmt_string(self, focused, active, value_key='msg', **kwargs):
        color_key, font_key = self.get_fmt_keys(focused, active, **kwargs)
        return '{{!%({})s,%({})s!}}%({})s{{!%({})s!}}'.format(
            color_key,
            font_key,
            value_key,
            'color_end',
        )

    def depend_skip(self):
        return False

    def has_input(self):
        return True

    @overrides(InputKeyHandler)
    def handle_read(self, c):
        return util.ReadState.IGNORED

    def render(self, screen, row, **kwargs):
        return 0

    @property
    def height(self):
        return 1

    def set_value(self, value):
        self.value = value

    def get_value(self):
        return self.value


class NoInputField(BaseField):
    @overrides(BaseField)
    def has_input(self):
        return False


class InputField(BaseField):
    def __init__(self, parent, name, message, format_default=None, **kwargs):
        BaseField.__init__(self, parent=parent, name=name, **kwargs)
        self.format_default = format_default
        self.message = None
        self.set_message(message)

    depend = None

    @overrides(BaseField)
    def handle_read(self, c):
        if c in [curses.KEY_ENTER, util.KEY_ENTER2, util.KEY_BACKSPACE2, 113]:
            return util.ReadState.READ
        return util.ReadState.IGNORED

    def set_message(self, msg):
        changed = self.message != msg
        self.message = msg
        return changed

    def set_depend(self, i, inverse=False):
        if not isinstance(i, CheckedInput):
            raise Exception('Can only depend on CheckedInputs')
        self.depend = i
        self.inverse = inverse

    def depend_skip(self):
        if not self.depend:
            return False
        if self.inverse:
            return self.depend.checked
        else:
            return not self.depend.checked


class Header(NoInputField):
    def __init__(self, parent, header, space_above, space_below, **kwargs):
        if 'name' not in kwargs:
            kwargs['name'] = header
        NoInputField.__init__(self, parent=parent, **kwargs)
        self.header = '{!white,black,bold!}%s' % header
        self.space_above = space_above
        self.space_below = space_below

    @overrides(BaseField)
    def render(self, screen, row, col=0, **kwargs):
        rows = 1
        if self.space_above:
            row += 1
            rows += 1
        self.parent.add_string(row, self.header, scr=screen, col=col, pad=False)
        if self.space_below:
            rows += 1
        return rows

    @property
    def height(self):
        return 1 + int(self.space_above) + int(self.space_below)


class InfoField(NoInputField):
    def __init__(self, parent, name, label, value, **kwargs):
        NoInputField.__init__(self, parent=parent, name=name, **kwargs)
        self.label = label
        self.value = value
        self.txt = f'{label} {value}'

    @overrides(BaseField)
    def render(self, screen, row, col=0, **kwargs):
        self.parent.add_string(row, self.txt, scr=screen, col=col, pad=False)
        return 1

    @overrides(BaseField)
    def set_value(self, v):
        self.value = v
        if isinstance(v, float):
            self.txt = f'{self.label} {self.value:.2f}'
        else:
            self.txt = f'{self.label} {self.value}'


class CheckedInput(InputField):
    def __init__(
        self,
        parent,
        name,
        message,
        checked=False,
        checked_char='X',
        unchecked_char=' ',
        checkbox_format='[%s] ',
        **kwargs,
    ):
        InputField.__init__(self, parent, name, message, **kwargs)
        self.set_value(checked)
        self.fmt_keys.update(
            {
                'msg': message,
                'checkbox_format': checkbox_format,
                'unchecked_char': unchecked_char,
                'checked_char': checked_char,
            }
        )
        self.set_fmt_key('font_checked', 'font', kwargs)
        self.set_fmt_key('font_unfocused_checked', 'font_checked', kwargs)
        self.set_fmt_key('font_active_checked', 'font_active', kwargs)
        self.set_fmt_key('font_unfocused_active_checked', 'font_active_checked', kwargs)
        self.set_fmt_key('color_checked', 'color', kwargs)
        self.set_fmt_key('color_active_checked', 'color_active', kwargs)
        self.set_fmt_key('color_unfocused_checked', 'color_checked', kwargs)
        self.set_fmt_key(
            'color_unfocused_active_checked', 'color_unfocused_active', kwargs
        )

    @property
    def checked(self):
        return self.value

    @overrides(BaseField)
    def get_fmt_keys(self, focused, active, **kwargs):
        color_key, font_key = super().get_fmt_keys(focused, active, **kwargs)
        if self.checked:
            color_key += '_checked'
            font_key += '_checked'
        return color_key, font_key

    def build_msg_string(self, focused, active):
        fmt_str = self.build_fmt_string(focused, active)
        char = self.fmt_keys['checked_char' if self.checked else 'unchecked_char']
        chk_box = ''
        try:
            chk_box = self.fmt_keys['checkbox_format'] % char
        except KeyError:
            pass
        msg = fmt_str % self.fmt_keys
        return chk_box + msg

    @overrides(InputField)
    def render(self, screen, row, col=0, **kwargs):
        string = self.build_msg_string(kwargs.get('focused'), kwargs.get('active'))

        self.parent.add_string(row, string, scr=screen, col=col, pad=False)
        return 1

    @overrides(InputField)
    def handle_read(self, c):
        if c == util.KEY_SPACE:
            self.set_value(not self.checked)
            return util.ReadState.CHANGED
        return util.ReadState.IGNORED

    @overrides(InputField)
    def set_message(self, msg):
        changed = InputField.set_message(self, msg)
        if 'msg' in self.fmt_keys and self.fmt_keys['msg'] != msg:
            changed = True
        self.fmt_keys.update({'msg': msg})

        return changed


class CheckedPlusInput(CheckedInput):
    def __init__(
        self,
        parent,
        name,
        message,
        child,
        child_always_visible=False,
        show_usage_hints=True,
        msg_fmt='%s ',
        **kwargs,
    ):
        CheckedInput.__init__(self, parent, name, message, **kwargs)
        self.child = child
        self.child_active = False
        self.show_usage_hints = show_usage_hints
        self.msg_fmt = msg_fmt
        self.child_always_visible = child_always_visible

    @property
    def height(self):
        return max(2 if self.show_usage_hints else 1, self.child.height)

    @overrides(CheckedInput)
    def render(
        self, screen, row, width=None, active=False, focused=False, col=0, **kwargs
    ):
        isact = active and not self.child_active
        CheckedInput.render(
            self, screen, row, width=width, active=isact, focused=focused, col=col
        )
        rows = 1
        if self.show_usage_hints and (
            self.child_always_visible or (active and self.checked)
        ):
            msg = '(esc to leave)' if self.child_active else '(right arrow to edit)'
            self.parent.add_string(row + 1, msg, scr=screen, col=col, pad=False)
            rows += 1

        msglen = len(
            self.msg_fmt % colors.strip_colors(self.build_msg_string(focused, active))
        )
        # show child
        if self.checked or self.child_always_visible:
            crows = self.child.render(
                screen,
                row,
                width=width - msglen,
                active=self.child_active and active,
                col=col + msglen,
                cursor_offset=msglen,
            )
            rows = max(rows, crows)
        else:
            self.parent.add_string(
                row,
                '(enable to view/edit value)',
                scr=screen,
                col=col + msglen,
                pad=False,
            )
        return rows

    @overrides(CheckedInput)
    def handle_read(self, c):
        if self.child_active:
            if c == util.KEY_ESC:  # leave child on esc
                self.child_active = False
                return util.ReadState.READ
            # pass keys through to child
            return self.child.handle_read(c)
        else:
            if c == util.KEY_SPACE:
                self.set_value(not self.checked)
                return util.ReadState.CHANGED
            if (self.checked or self.child_always_visible) and c == curses.KEY_RIGHT:
                self.child_active = True
                return util.ReadState.READ
        return util.ReadState.IGNORED

    def get_child(self):
        return self.child


class IntSpinInput(InputField):
    def __init__(
        self,
        parent,
        name,
        message,
        move_func,
        value,
        min_val=None,
        max_val=None,
        inc_amt=1,
        incr_large=10,
        strict_validation=False,
        fmt='%d',
        **kwargs,
    ):
        InputField.__init__(self, parent, name, message, **kwargs)
        self.convert_func = int
        self.fmt = fmt
        self.valstr = str(value)
        self.default_str = self.valstr
        self.set_value(value)
        self.default_value = self.value
        self.last_valid_value = self.value
        self.last_active = False
        self.cursor = len(self.valstr)
        self.cursoff = (
            colors.get_line_width(self.message) + 3
        )  # + 4 for the " [ " in the rendered string
        self.move_func = move_func
        self.strict_validation = strict_validation
        self.min_val = min_val
        self.max_val = max_val
        self.inc_amt = inc_amt
        self.incr_large = incr_large

    def validate_value(self, value, on_invalid=None):
        if (self.min_val is not None) and value < self.min_val:
            value = on_invalid if on_invalid else self.min_val
        if (self.max_val is not None) and value > self.max_val:
            value = on_invalid if on_invalid else self.max_val
        return value

    @overrides(InputField)
    def render(
        self, screen, row, active=False, focused=True, col=0, cursor_offset=0, **kwargs
    ):
        if active:
            self.last_active = True
        elif self.last_active:
            self.set_value(
                self.valstr, validate=True, value_on_fail=self.last_valid_value
            )
            self.last_active = False

        fmt_str = self.build_fmt_string(focused, active, value_key='value')
        value_format = '%(msg)s {!input!}'
        if not self.valstr:
            value_format += '[  ]'
        elif self.format_default and self.valstr == self.default_str:
            value_format += '[ {!magenta,black!}%(value)s{!input!} ]'
        else:
            value_format += '[ ' + fmt_str + ' ]'

        self.parent.add_string(
            row,
            value_format
            % dict({'msg': self.message, 'value': '%s' % self.valstr}, **self.fmt_keys),
            scr=screen,
            col=col,
            pad=False,
        )
        if active:
            if focused:
                util.safe_curs_set(util.Curser.NORMAL)
                self.move_func(row, self.cursor + self.cursoff + cursor_offset)
            else:
                util.safe_curs_set(util.Curser.INVISIBLE)
        return 1

    @overrides(InputField)
    def handle_read(self, c):
        if c == util.KEY_SPACE:
            return util.ReadState.READ
        elif c == curses.KEY_PPAGE:
            self.set_value(self.value + self.inc_amt, validate=True)
        elif c == curses.KEY_NPAGE:
            self.set_value(self.value - self.inc_amt, validate=True)
        elif c == util.KEY_ALT_AND_KEY_PPAGE:
            self.set_value(self.value + self.incr_large, validate=True)
        elif c == util.KEY_ALT_AND_KEY_NPAGE:
            self.set_value(self.value - self.incr_large, validate=True)
        elif c == curses.KEY_LEFT:
            self.cursor = max(0, self.cursor - 1)
        elif c == curses.KEY_RIGHT:
            self.cursor = min(len(self.valstr), self.cursor + 1)
        elif c == curses.KEY_HOME:
            self.cursor = 0
        elif c == curses.KEY_END:
            self.cursor = len(self.valstr)
        elif c == curses.KEY_BACKSPACE or c == util.KEY_BACKSPACE2:
            if self.valstr and self.cursor > 0:
                new_val = self.valstr[: self.cursor - 1] + self.valstr[self.cursor :]
                self.set_value(
                    new_val,
                    validate=False,
                    cursor=self.cursor - 1,
                    cursor_on_fail=True,
                    value_on_fail=self.valstr if self.strict_validation else None,
                )
        elif c == curses.KEY_DC:  # Del
            if self.valstr and self.cursor <= len(self.valstr):
                if self.cursor == 0:
                    new_val = self.valstr[1:]
                else:
                    new_val = (
                        self.valstr[: self.cursor] + self.valstr[self.cursor + 1 :]
                    )
                self.set_value(
                    new_val,
                    validate=False,
                    cursor=False,
                    value_on_fail=self.valstr if self.strict_validation else None,
                    cursor_on_fail=True,
                )
        elif c == ord('-'):  # minus
            self.set_value(
                self.value - 1,
                validate=True,
                cursor=True,
                cursor_on_fail=True,
                value_on_fail=self.value,
                on_invalid=self.value,
            )
        elif c == ord('+'):  # plus
            self.set_value(
                self.value + 1,
                validate=True,
                cursor=True,
                cursor_on_fail=True,
                value_on_fail=self.value,
                on_invalid=self.value,
            )
        elif util.is_int_chr(c):
            if self.strict_validation:
                new_val = (
                    self.valstr[: self.cursor - 1]
                    + chr(c)
                    + self.valstr[self.cursor - 1 :]
                )
                self.set_value(
                    new_val,
                    validate=True,
                    cursor=self.cursor + 1,
                    value_on_fail=self.valstr,
                    on_invalid=self.value,
                )
            else:
                minus_place = self.valstr.find('-')
                if self.cursor > minus_place:
                    new_val = (
                        self.valstr[: self.cursor] + chr(c) + self.valstr[self.cursor :]
                    )
                    self.set_value(
                        new_val,
                        validate=True,
                        cursor=self.cursor + 1,
                        on_invalid=self.value,
                    )
        else:
            return util.ReadState.IGNORED
        return util.ReadState.READ

    @overrides(BaseField)
    def set_value(
        self,
        val,
        cursor=True,
        validate=False,
        cursor_on_fail=False,
        value_on_fail=None,
        on_invalid=None,
    ):
        value = None
        try:
            value = self.convert_func(val)
            if validate:
                validated = self.validate_value(value, on_invalid)
                if validated != value:
                    # Value was not valid, so use validated value instead.
                    # Also set cursor according to validated value
                    cursor = True
                    value = validated

            new_valstr = self.fmt % value
            if new_valstr == self.valstr:
                # If string has not change, keep cursor
                cursor = False
            self.valstr = new_valstr
            self.last_valid_value = self.value = value
        except ValueError:
            if value_on_fail is not None:
                self.set_value(
                    value_on_fail,
                    cursor=cursor,
                    cursor_on_fail=cursor_on_fail,
                    validate=validate,
                    on_invalid=on_invalid,
                )
                return
            self.value = None
            self.valstr = val
            if cursor_on_fail:
                self.cursor = cursor
        except TypeError:
            import traceback

            log.warning('TypeError: %s', ''.join(traceback.format_exc()))
        else:
            if cursor is True:
                self.cursor = len(self.valstr)
            elif cursor is not False:
                self.cursor = cursor


class FloatSpinInput(IntSpinInput):
    def __init__(self, parent, message, name, move_func, value, precision=1, **kwargs):
        self.precision = precision
        IntSpinInput.__init__(self, parent, message, name, move_func, value, **kwargs)
        self.fmt = '%%.%df' % precision
        self.convert_func = lambda valstr: round(float(valstr), self.precision)
        self.set_value(value)
        self.cursor = len(self.valstr)

    @overrides(IntSpinInput)
    def handle_read(self, c):
        if c == ord('.'):
            minus_place = self.valstr.find('-')
            if self.cursor <= minus_place:
                return util.ReadState.READ
            point_place = self.valstr.find('.')
            if point_place >= 0:
                return util.ReadState.READ
            new_val = self.valstr[: self.cursor] + chr(c) + self.valstr[self.cursor :]
            self.set_value(new_val, validate=True, cursor=self.cursor + 1)
        else:
            return IntSpinInput.handle_read(self, c)


class SelectInput(InputField):
    def __init__(
        self,
        parent,
        name,
        message,
        opts,
        vals,
        active_index,
        active_default=False,
        require_select_action=True,
        **kwargs,
    ):
        InputField.__init__(self, parent, name, message, **kwargs)
        self.opts = opts
        self.vals = vals
        self.active_index = active_index
        self.selected_index = active_index
        self.default_option = active_index if active_default else None
        self.require_select_action = require_select_action
        self.fmt_keys.update({'font_active': 'bold'})
        font_selected = kwargs.get('font_selected', 'bold,underline')

        self.set_fmt_key('font_selected', font_selected, kwargs)
        self.set_fmt_key('font_active_selected', 'font_selected', kwargs)
        self.set_fmt_key('font_unfocused_selected', 'font_selected', kwargs)
        self.set_fmt_key(
            'font_unfocused_active_selected', 'font_active_selected', kwargs
        )

        self.set_fmt_key('color_selected', 'color', kwargs)
        self.set_fmt_key('color_active_selected', 'color_active', kwargs)
        self.set_fmt_key('color_unfocused_selected', 'color_selected', kwargs)
        self.set_fmt_key(
            'color_unfocused_active_selected', 'color_unfocused_active', kwargs
        )
        self.set_fmt_key('color_default_value', 'magenta,black', kwargs)

        self.set_fmt_key('color_default_value', 'magenta,black')
        self.set_fmt_key('color_default_value_active', 'magentadark,white')
        self.set_fmt_key('color_default_value_selected', 'color_default_value', kwargs)
        self.set_fmt_key('color_default_value_unfocused', 'color_default_value', kwargs)
        self.set_fmt_key(
            'color_default_value_unfocused_selected',
            'color_default_value_selected',
            kwargs,
        )
        self.set_fmt_key('color_default_value_active_selected', 'magentadark,white')
        self.set_fmt_key(
            'color_default_value_unfocused_active_selected',
            'color_unfocused_active',
            kwargs,
        )

    @property
    def height(self):
        return 1 + bool(self.message)

    @overrides(BaseField)
    def get_fmt_keys(self, focused, active, selected=False, **kwargs):
        color_key, font_key = super().get_fmt_keys(focused, active, **kwargs)
        if selected:
            color_key += '_selected'
            font_key += '_selected'
        return color_key, font_key

    @overrides(InputField)
    def render(self, screen, row, active=False, focused=True, col=0, **kwargs):
        if self.message:
            self.parent.add_string(row, self.message, scr=screen, col=col, pad=False)
            row += 1

        off = col + 1
        for i, opt in enumerate(self.opts):
            self.fmt_keys['msg'] = opt
            fmt_args = {'selected': i == self.selected_index}
            if i == self.default_option:
                fmt_args['color_key'] = 'color_default_value'
            fmt = self.build_fmt_string(
                focused, (i == self.active_index) and active, **fmt_args
            )
            string = '[%s]' % (fmt % self.fmt_keys)
            self.parent.add_string(row, string, scr=screen, col=off, pad=False)
            off += len(opt) + 3
        if self.message:
            return 2
        else:
            return 1

    @overrides(InputField)
    def handle_read(self, c):
        if c == curses.KEY_LEFT:
            self.active_index = max(0, self.active_index - 1)
            if not self.require_select_action:
                self.selected_index = self.active_index
        elif c == curses.KEY_RIGHT:
            self.active_index = min(len(self.opts) - 1, self.active_index + 1)
            if not self.require_select_action:
                self.selected_index = self.active_index
        elif c == ord(' '):
            if self.require_select_action:
                self.selected_index = self.active_index
        else:
            return util.ReadState.IGNORED
        return util.ReadState.READ

    @overrides(BaseField)
    def get_value(self):
        return self.vals[self.selected_index]

    @overrides(BaseField)
    def set_value(self, value):
        for i, val in enumerate(self.vals):
            if value == val:
                self.selected_index = i
                return
        raise Exception('Invalid value for SelectInput')


class TextInput(InputField):
    def __init__(
        self,
        parent,
        name,
        message,
        move_func,
        width,
        value,
        complete=False,
        activate_input=False,
        **kwargs,
    ):
        InputField.__init__(self, parent, name, message, **kwargs)
        self.move_func = move_func
        self._width = width
        self.value = value if value else ''
        self.default_value = value
        self.complete = complete
        self.tab_count = 0
        self.cursor = len(self.value)
        self.opts = None
        self.opt_off = 0
        self.value_offset = 0
        self.activate_input = activate_input  # Wether input must be activated
        self.input_active = not self.activate_input

    @property
    def width(self):
        return self._width

    @property
    def height(self):
        return 1 + bool(self.message)

    def calculate_textfield_value(self, width, cursor_offset):
        cursor_width = width

        if self.cursor > (cursor_width - 1):
            c_pos_abs = self.cursor - cursor_width
            if cursor_width <= (self.cursor - self.value_offset):
                new_cur = c_pos_abs + 1
                self.value_offset = new_cur
            else:
                if self.cursor >= len(self.value):
                    c_pos_abs = len(self.value) - cursor_width
                    new_cur = c_pos_abs + 1
                    self.value_offset = new_cur
            vstr = self.value[self.value_offset :]

            if len(vstr) > cursor_width:
                vstr = vstr[:cursor_width]
            vstr = vstr.ljust(cursor_width)
        else:
            if len(self.value) <= cursor_width:
                self.value_offset = 0
                vstr = self.value.ljust(cursor_width)
            else:
                self.value_offset = min(self.value_offset, self.cursor)
                vstr = self.value[self.value_offset :]
                if len(vstr) > cursor_width:
                    vstr = vstr[:cursor_width]
            vstr = vstr.ljust(cursor_width)

        return vstr

    def calculate_cursor_pos(self, width, col):
        cursor_width = width
        x_pos = self.cursor + col

        if (self.cursor + col - self.value_offset) > cursor_width:
            x_pos += self.value_offset
        else:
            x_pos -= self.value_offset

        return min(width - 1 + col, x_pos)

    @overrides(InputField)
    def render(
        self,
        screen,
        row,
        width=None,
        active=False,
        focused=True,
        col=0,
        cursor_offset=0,
        **kwargs,
    ):
        if not self.value and not active and len(self.default_value) != 0:
            self.value = self.default_value
            self.cursor = len(self.value)

        if self.message:
            self.parent.add_string(row, self.message, scr=screen, col=col, pad=False)
            row += 1

        vstr = self.calculate_textfield_value(width, cursor_offset)

        if active:
            if self.opts:
                self.parent.add_string(
                    row + 1, self.opts[self.opt_off :], scr=screen, col=col, pad=False
                )

            if focused and self.input_active:
                util.safe_curs_set(
                    util.Curser.NORMAL
                )  # Make cursor visible when text field is focused
                x_pos = self.calculate_cursor_pos(width, col)
                self.move_func(row, x_pos)

        fmt = '{!black,white,bold!}%s'
        if (
            self.format_default
            and len(self.value) != 0
            and self.value == self.default_value
        ):
            fmt = '{!magenta,white!}%s'
        if not active or not focused or self.input_active:
            fmt = '{!white,grey,bold!}%s'

        self.parent.add_string(
            row, fmt % vstr, scr=screen, col=col, pad=False, trim=False
        )
        return self.height

    @overrides(BaseField)
    def set_value(self, val):
        self.value = val
        self.cursor = len(self.value)

    @overrides(InputField)
    def handle_read(self, c):
        """
        Return False when key was swallowed, i.e. we recognised
        the key and no further action by other components should
        be performed.
        """
        if self.activate_input:
            if not self.input_active:
                if c in [
                    curses.KEY_LEFT,
                    curses.KEY_RIGHT,
                    curses.KEY_HOME,
                    curses.KEY_END,
                    curses.KEY_ENTER,
                    util.KEY_ENTER2,
                ]:
                    self.input_active = True
                    return util.ReadState.READ
                else:
                    return util.ReadState.IGNORED
            elif c == util.KEY_ESC:
                self.input_active = False
                return util.ReadState.READ

        if c == util.KEY_TAB and self.complete:
            # Keep track of tab hit count to know when it's double-hit
            self.tab_count += 1
            if self.tab_count > 1:
                second_hit = True
                self.tab_count = 0
            else:
                second_hit = False

            # We only call the tab completer function if we're at the end of
            # the input string on the cursor is on a space
            if self.cursor == len(self.value) or self.value[self.cursor] == ' ':
                if self.opts:
                    prev = self.opt_off
                    self.opt_off += self.width - 3
                    # now find previous double space, best guess at a split point
                    # in future could keep opts unjoined to get this really right
                    self.opt_off = self.opts.rfind('  ', 0, self.opt_off) + 2
                    if (
                        second_hit and self.opt_off == prev
                    ):  # double tap and we're at the end
                        self.opt_off = 0
                else:
                    opts = self.do_complete(self.value)
                    if len(opts) == 1:  # only one option, just complete it
                        self.value = opts[0]
                        self.cursor = len(opts[0])
                        self.tab_count = 0
                    elif len(opts) > 1:
                        prefix = os.path.commonprefix(opts)
                        if prefix:
                            self.value = prefix
                            self.cursor = len(prefix)

                    if (
                        len(opts) > 1 and second_hit
                    ):  # display multiple options on second tab hit
                        sp = self.value.rfind(os.sep) + 1
                        self.opts = '  '.join([o[sp:] for o in opts])

        # Cursor movement
        elif c == curses.KEY_LEFT:
            self.cursor = max(0, self.cursor - 1)
        elif c == curses.KEY_RIGHT:
            self.cursor = min(len(self.value), self.cursor + 1)
        elif c == curses.KEY_HOME:
            self.cursor = 0
        elif c == curses.KEY_END:
            self.cursor = len(self.value)

        # Delete a character in the input string based on cursor position
        elif c == curses.KEY_BACKSPACE or c == util.KEY_BACKSPACE2:
            if self.value and self.cursor > 0:
                self.value = self.value[: self.cursor - 1] + self.value[self.cursor :]
                self.cursor -= 1
        elif c == [util.KEY_ESC, util.KEY_BACKSPACE2] or c == [
            util.KEY_ESC,
            curses.KEY_BACKSPACE,
        ]:
            self.value, self.cursor = delete_alt_backspace(self.value, self.cursor)
        elif c == curses.KEY_DC:
            if self.value and self.cursor < len(self.value):
                self.value = self.value[: self.cursor] + self.value[self.cursor + 1 :]
        elif c > 31 and c < 256:
            # Emulate getwch
            stroke = chr(c)
            uchar = stroke
            while not uchar:
                try:
                    uchar = stroke.decode(self.parent.encoding)
                except UnicodeDecodeError:
                    c = self.parent.parent.stdscr.getch()
                    stroke += chr(c)
            if uchar:
                if self.cursor == len(self.value):
                    self.value += uchar
                else:
                    # Insert into string
                    self.value = (
                        self.value[: self.cursor] + uchar + self.value[self.cursor :]
                    )
                # Move the cursor forward
                self.cursor += 1

        else:
            self.opts = None
            self.opt_off = 0
            self.tab_count = 0
            return util.ReadState.IGNORED
        return util.ReadState.READ

    def do_complete(self, line):
        line = os.path.abspath(os.path.expanduser(line))
        ret = []
        if os.path.exists(line):
            # This is a correct path, check to see if it's a directory
            if os.path.isdir(line):
                # Directory, so we need to show contents of directory
                for f in os.listdir(line):
                    # Skip hidden
                    if f.startswith('.'):
                        continue
                    f = os.path.join(line, f)
                    if os.path.isdir(f):
                        f += os.sep
                    ret.append(f)
            else:
                # This is a file, but we could be looking for another file that
                # shares a common prefix.
                for f in os.listdir(os.path.dirname(line)):
                    if f.startswith(os.path.split(line)[1]):
                        ret.append(os.path.join(os.path.dirname(line), f))
        else:
            # This path does not exist, so lets do a listdir on it's parent
            # and find any matches.
            ret = []
            if os.path.isdir(os.path.dirname(line)):
                for f in os.listdir(os.path.dirname(line)):
                    if f.startswith(os.path.split(line)[1]):
                        p = os.path.join(os.path.dirname(line), f)

                        if os.path.isdir(p):
                            p += os.sep
                        ret.append(p)
        return ret


class ComboInput(InputField):
    def __init__(
        self, parent, name, message, choices, default=None, searchable=True, **kwargs
    ):
        InputField.__init__(self, parent, name, message, **kwargs)
        self.choices = choices
        self.default = default
        self.set_value(default)
        max_width = 0
        for c in choices:
            max_width = max(max_width, len(c[1]))
        self.choices_width = max_width
        self.searchable = searchable

    @overrides(BaseField)
    def render(self, screen, row, col=0, **kwargs):
        fmt_str = self.build_fmt_string(kwargs.get('focused'), kwargs.get('active'))
        string = '%s: [%10s]' % (self.message, fmt_str % self.fmt_keys)
        self.parent.add_string(row, string, scr=screen, col=col, pad=False)
        return 1

    def _lang_selected(self, selected, *args, **kwargs):
        if selected is not None:
            self.set_value(selected)
        self.parent.pop_popup()

    @overrides(InputField)
    def handle_read(self, c):
        if c in [util.KEY_SPACE, curses.KEY_ENTER, util.KEY_ENTER2]:

            def search_handler(key):
                """Handle keyboard input to seach the list"""
                if not util.is_printable_chr(key):
                    return
                selected = select_popup.current_selection()

                def select_in_range(begin, end):
                    for i in range(begin, end):
                        val = select_popup.inputs[i].get_value()
                        if val.lower().startswith(chr(key)):
                            select_popup.set_selection(i)
                            return True
                    return False

                # First search downwards
                if not select_in_range(selected + 1, len(select_popup.inputs)):
                    # No match, so start at beginning
                    select_in_range(0, selected)

            from deluge.ui.console.widgets.popup import (  # Must import here
                SelectablePopup,
            )

            select_popup = SelectablePopup(
                self.parent,
                ' %s ' % _('Select Language'),
                self._lang_selected,
                input_cb=search_handler if self.searchable else None,
                border_off_west=1,
                active_wrap=False,
                width_req=self.choices_width + 12,
            )
            for choice in self.choices:
                args = {'data': choice[0]}
                select_popup.add_line(
                    choice[0],
                    choice[1],
                    selectable=True,
                    selected=choice[0] == self.get_value(),
                    **args,
                )
            self.parent.push_popup(select_popup)
            return util.ReadState.CHANGED
        return util.ReadState.IGNORED

    @overrides(BaseField)
    def set_value(self, val):
        self.value = val
        msg = None
        for c in self.choices:
            if c[0] == val:
                msg = c[1]
                break
        if msg is None:
            log.warning(
                'Setting value "%s" found nothing in choices: %s', val, self.choices
            )
        self.fmt_keys.update({'msg': msg})


class TextField(BaseField):
    def __init__(self, parent, name, value, selectable=True, value_fmt='%s', **kwargs):
        BaseField.__init__(
            self, parent=parent, name=name, selectable=selectable, **kwargs
        )
        self.value = value
        self.value_fmt = value_fmt
        self.set_value(value)

    @overrides(BaseField)
    def set_value(self, value):
        self.value = value
        self.txt = self.value_fmt % (value)

    @overrides(BaseField)
    def has_input(self):
        return True

    @overrides(BaseField)
    def render(self, screen, row, active=False, focused=False, col=0, **kwargs):
        util.safe_curs_set(
            util.Curser.INVISIBLE
        )  # Make cursor invisible when text field is active
        fmt = self.build_fmt_string(focused, active)
        self.fmt_keys['msg'] = self.txt
        string = fmt % self.fmt_keys
        self.parent.add_string(row, string, scr=screen, col=col, pad=False, trim=False)
        return 1


class TextArea(TextField):
    def __init__(self, parent, name, value, value_fmt='%s', **kwargs):
        TextField.__init__(
            self, parent, name, value, selectable=False, value_fmt=value_fmt, **kwargs
        )

    @overrides(TextField)
    def render(self, screen, row, col=0, **kwargs):
        util.safe_curs_set(
            util.Curser.INVISIBLE
        )  # Make cursor invisible when text field is active
        color = '{!white,black!}'
        lines = wrap_string(self.txt, self.parent.width - 3, 3, True)

        for i, line in enumerate(lines):
            self.parent.add_string(
                row + i,
                f'{color}{line}',
                scr=screen,
                col=col,
                pad=False,
                trim=False,
            )
        return len(lines)

    @property
    def height(self):
        lines = wrap_string(self.txt, self.parent.width - 3, 3, True)
        return len(lines)

    @overrides(TextField)
    def has_input(self):
        return False


class DividerField(NoInputField):
    def __init__(
        self,
        parent,
        name,
        value,
        selectable=False,
        fill_width=True,
        value_fmt='%s',
        **kwargs,
    ):
        NoInputField.__init__(
            self, parent=parent, name=name, selectable=selectable, **kwargs
        )
        self.value = value
        self.value_fmt = value_fmt
        self.set_value(value)
        self.fill_width = fill_width

    @overrides(BaseField)
    def set_value(self, value):
        self.value = value
        self.txt = self.value_fmt % (value)

    @overrides(BaseField)
    def render(
        self, screen, row, active=False, focused=False, col=0, width=None, **kwargs
    ):
        util.safe_curs_set(
            util.Curser.INVISIBLE
        )  # Make cursor invisible when text field is active
        fmt = self.build_fmt_string(focused, active)
        self.fmt_keys['msg'] = self.txt
        if self.fill_width:
            self.fmt_keys['msg'] = ''
            string_len = len(remove_formatting(fmt % self.fmt_keys))
            fill_len = width - string_len - (len(self.txt) - 1)
            self.fmt_keys['msg'] = self.txt * fill_len
        string = fmt % self.fmt_keys
        self.parent.add_string(row, string, scr=screen, col=col, pad=False, trim=False)
        return 1



#
# Copyright (C) 2009-2010 John Garland <johnnybg+deluge@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#
# pylint: disable=redefined-builtin

import bz2
import gzip
import zipfile


def Zipped(reader):  # NOQA: N802
    """Blocklist reader for zipped blocklists"""

    def _open(self):
        z = zipfile.ZipFile(self.file)
        f = z.open(z.namelist()[0])
        return f

    reader.open = _open
    return reader


def GZipped(reader):  # NOQA: N802
    """Blocklist reader for gzipped blocklists"""

    def _open(self):
        return gzip.open(self.file)

    reader.open = _open
    return reader


def BZipped2(reader):  # NOQA: N802
    """Blocklist reader for bzipped2 blocklists"""

    def _open(self):
        return bz2.BZ2File(self.file)

    reader.open = _open
    return reader



#
# Copyright (C) 2009 Damien Churchill <damoxc@gmail.com>
#
# Basic plugin template created by:
# Copyright (C) 2008 Martijn Voncken <mvoncken@gmail.com>
# Copyright (C) 2007-2009 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging

from twisted.internet import defer
from twisted.internet.error import CannotListenError

import deluge.component as component
from deluge import configmanager
from deluge.core.rpcserver import export
from deluge.plugins.pluginbase import CorePluginBase

try:
    from deluge.ui.web import server
except ImportError:
    server = False

log = logging.getLogger(__name__)

DEFAULT_PREFS = {'enabled': False, 'ssl': False, 'port': 8112}


class Core(CorePluginBase):
    server = None

    def enable(self):
        self.config = configmanager.ConfigManager('web_plugin.conf', DEFAULT_PREFS)
        if self.config['enabled']:
            self.start_server()

    def disable(self):
        self.stop_server()

    def update(self):
        pass

    def _on_stop(self, *args):
        return self.start_server()

    @export
    def got_deluge_web(self):
        """Status of deluge-web module installation.

        Check if deluge.ui.web.server modulge is installed and has been successfully imported.

        Returns:
            bool: True is deluge-web is installed and available, otherwise False.

        """

        return bool(server)

    def start_server(self):
        if not self.server:
            if not self.got_deluge_web():
                return False

            try:
                self.server = component.get('DelugeWeb')
            except KeyError:
                self.server = server.DelugeWeb(daemon=False)

        self.server.port = self.config['port']
        self.server.https = self.config['ssl']
        try:
            self.server.start()
        except CannotListenError as ex:
            log.warning('Failed to start WebUI server: %s', ex)
            raise
        return True

    def stop_server(self):
        if self.server:
            return self.server.stop()
        return defer.succeed(True)

    def restart_server(self):
        return self.stop_server().addCallback(self._on_stop)

    @export
    def set_config(self, config):
        """Sets the config dictionary."""

        action = None
        if 'enabled' in config:
            if config['enabled'] != self.config['enabled']:
                action = config['enabled'] and 'start' or 'stop'

        if 'ssl' in config:
            if not action:
                action = 'restart'

        for key in config:
            self.config[key] = config[key]
        self.config.save()

        if action == 'start':
            return self.start_server()
        elif action == 'stop':
            return self.stop_server()
        elif action == 'restart':
            return self.restart_server()

    @export
    def get_config(self):
        """Returns the config dictionary."""
        return self.config.config



#
# Copyright (C) 2010 John Garland <johnnybg+deluge@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging
import os
import tempfile
from html.parser import HTMLParser
from urllib.parse import urljoin, urlparse

from twisted.internet import defer, threads
from twisted.python.failure import Failure
from twisted.web.resource import ForbiddenResource, NoResource

from deluge.component import Component
from deluge.configmanager import get_config_dir
from deluge.decorators import proxy
from deluge.httpdownloader import download_file

try:
    import chardet
except ImportError:
    chardet = None

try:
    from PIL import Image
except ImportError:
    Image = None

log = logging.getLogger(__name__)


class TrackerIcon:
    """
    Represents a tracker's icon
    """

    def __init__(self, filename):
        """
        Initialises a new TrackerIcon object

        :param filename: the filename of the icon
        :type filename: string
        """
        self.filename = os.path.abspath(filename)
        self.mimetype = extension_to_mimetype(self.filename.rpartition('.')[2])
        self.data = None
        self.icon_cache = None

    def __eq__(self, other):
        """
        Compares this TrackerIcon with another to determine if they're equal

        :param other: the TrackerIcon to compare to
        :type other: TrackerIcon
        :returns: whether or not they're equal
        :rtype: boolean
        """
        return (
            os.path.samefile(self.filename, other.filename)
            or self.get_mimetype() == other.get_mimetype()
            and self.get_data() == other.get_data()
        )

    def get_mimetype(self):
        """
        Returns the mimetype of this TrackerIcon's image

        :returns: the mimetype of the image
        :rtype: string
        """
        return self.mimetype

    def get_data(self):
        """
        Returns the TrackerIcon's image data as a string

        :returns: the image data
        :rtype: string
        """
        if not self.data:
            with open(self.filename, 'rb') as _file:
                self.data = _file.read()
        return self.data

    def get_filename(self, full=True):
        """
        Returns the TrackerIcon image's filename

        :param full: an (optional) arg to indicate whether or not to
                     return the full path
        :type full: boolean
        :returns: the path of the TrackerIcon's image
        :rtype: string
        """
        return self.filename if full else os.path.basename(self.filename)

    def set_cached_icon(self, data):
        """
        Set the cached icon data.

        """
        self.icon_cache = data

    def get_cached_icon(self):
        """
        Returns the cached icon data.

        """
        return self.icon_cache


class TrackerIcons(Component):
    """
    A TrackerIcon factory class
    """

    def __init__(self, icon_dir=None, no_icon=None):
        """
        Initialises a new TrackerIcons object

        :param icon_dir: the (optional) directory of where to store the icons
        :type icon_dir: string
        :param no_icon: the (optional) path name of the icon to show when no icon
                       can be fetched
        :type no_icon: string
        """
        Component.__init__(self, 'TrackerIcons')
        if not icon_dir:
            icon_dir = get_config_dir('icons')
        self.dir = icon_dir
        if not os.path.isdir(self.dir):
            os.makedirs(self.dir)

        self.icons = {}
        for icon in os.listdir(self.dir):
            if icon != no_icon:
                host = icon_name_to_host(icon)
                try:
                    self.icons[host] = TrackerIcon(os.path.join(self.dir, icon))
                except KeyError:
                    log.warning('invalid icon %s', icon)
        if no_icon:
            self.icons[None] = TrackerIcon(no_icon)
        else:
            self.icons[None] = None
        self.icons[''] = self.icons[None]

        self.pending = {}
        self.redirects = {}

    def has(self, host):
        """
        Returns True or False if the tracker icon for the given host exists or not.

        :param host: the host for the TrackerIcon
        :type host: string
        :returns: True or False
        :rtype: bool
        """
        return host.lower() in self.icons

    def get(self, host):
        """
        Returns a TrackerIcon for the given tracker's host
        from the icon cache.

        :param host: the host for the TrackerIcon
        :type host: string
        :returns: the TrackerIcon for the host
        :rtype: TrackerIcon
        """
        host = host.lower()
        if host in self.icons:
            return self.icons[host]
        else:
            return None

    def fetch(self, host):
        """
        Fetches (downloads) the icon for the given host.
        When the icon is downloaded a callback is fired
        on the the queue of callers to this function.

        :param host: the host to obtain the TrackerIcon for
        :type host: string
        :returns: a Deferred which fires with the TrackerIcon for the given host
        :rtype: Deferred
        """
        host = host.lower()
        if host in self.icons:
            # We already have it, so let's return it
            d = defer.succeed(self.icons[host])
        elif host in self.pending:
            # We're in the middle of getting it
            # Add ourselves to the waiting list
            d = defer.Deferred()
            self.pending[host].append(d)
        else:
            # We need to fetch it
            self.pending[host] = []
            tmp_file = tempfile.mkstemp(prefix='deluge_trackericon_html.')
            filename = tmp_file[1]
            # Start callback chain
            d = self.download_page(host, filename)
            d.addCallbacks(
                self.on_download_page_complete,
                self.on_download_page_fail,
            )
            d.addCallback(self.parse_html_page)
            d.addCallbacks(
                self.on_parse_complete, self.on_parse_fail, callbackArgs=(host,)
            )
            d.addBoth(self.del_tmp_file, tmp_file)
            d.addCallback(self.download_icon, host)
            d.addCallbacks(
                self.on_download_icon_complete,
                self.on_download_icon_fail,
                callbackArgs=(host,),
                errbackArgs=(host,),
            )
            d.addCallback(self.resize_icon)
            d.addCallback(self.store_icon, host)
        return d

    @staticmethod
    def del_tmp_file(result, tmp_file):
        """Remove tmp_file created when downloading tracker page"""
        fd, filename = tmp_file
        try:
            os.close(fd)
            os.remove(filename)
        except OSError:
            log.debug(f'Unable to delete temporary file: {filename}')

        return result

    def download_page(
        self, host: str, filename: str, url: str = None
    ) -> 'defer.Deferred[str]':
        """Downloads a tracker host's page

        If no url is provided, it bases the url on the host

        Args:
            host: The tracker host
            filename: Location to download page
            url: The url of the host

        Returns:
            The filename of the tracker host's page
        """
        if not url:
            url = self.host_to_url(host)

        log.debug(f'Downloading {host} {url} to {filename}')
        return download_file(url, filename, force_filename=True)

    def on_download_page_complete(self, page):
        """
        Runs any download clean up functions

        :param page: the page that finished downloading
        :type page: string
        :returns: the page that finished downloading
        :rtype: string
        """
        log.debug('Finished downloading %s', page)
        return page

    def on_download_page_fail(self, failure: 'Failure') -> 'Failure':
        """Runs any download failure clean-up functions

        Args:
            failure: The failure that occurred.

        Returns:
            The original failure.

        """
        log.debug(f'Error downloading page: {failure.getErrorMessage()}')
        return failure

    @proxy(threads.deferToThread)
    def parse_html_page(self, page):
        """
        Parses the html page for favicons

        :param page: the page to parse
        :type page: string
        :returns: a Deferred which callbacks a list of available favicons (url, type)
        :rtype: Deferred
        """
        encoding = 'UTF-8'
        if chardet:
            with open(page, 'rb') as _file:
                result = chardet.detect(_file.read())
                encoding = result['encoding']

        with open(page, encoding=encoding) as _file:
            parser = FaviconParser()
            for line in _file:
                parser.feed(line)
                if parser.left_head:
                    break
            parser.close()

        return parser.get_icons()

    def on_parse_complete(self, icons, host):
        """
        Runs any parse clean up functions

        :param icons: the icons that were extracted from the page
        :type icons: list
        :param host: the host the icons are for
        :type host: string
        :returns: the icons that were extracted from the page
        :rtype: list
        """
        log.debug('Parse Complete, got icons for %s: %s', host, icons)
        url = self.host_to_url(host)
        icons = [(urljoin(url, icon), mimetype) for icon, mimetype in icons]
        log.debug('Icon urls from %s: %s', host, icons)
        return icons

    def on_parse_fail(self, f):
        """
        Recovers from a parse error

        :param f: the failure that occurred
        :type f: Failure
        :returns: a Deferred if recovery was possible
                  else the original failure
        :rtype: Deferred or Failure
        """
        log.debug('Error parsing page: %s', f.getErrorMessage())
        return f

    def download_icon(self, icons, host):
        """
        Downloads the first available icon from icons

        :param icons: a list of icons
        :type icons: list
        :param host: the tracker's host name
        :type host: string
        :returns: a Deferred which fires with the downloaded icon's filename
        :rtype: Deferred
        """
        if len(icons) == 0:
            raise NoIconsError('empty icons list')
        (url, mimetype) = icons.pop(0)
        d = download_file(
            url,
            os.path.join(self.dir, host_to_icon_name(host, mimetype)),
            force_filename=True,
        )
        d.addCallback(self.check_icon_is_valid)
        if icons:
            d.addErrback(self.on_download_icon_fail, host, icons)
        return d

    @proxy(threads.deferToThread)
    def check_icon_is_valid(self, icon_name):
        """
        Performs a sanity check on icon_name

        :param icon_name: the name of the icon to check
        :type icon_name: string
        :returns: the name of the validated icon
        :rtype: string
        :raises: InvalidIconError
        """

        if Image:
            try:
                with Image.open(icon_name):
                    pass
            except OSError as ex:
                raise InvalidIconError(ex)
        else:
            if not os.path.getsize(icon_name):
                raise InvalidIconError('empty icon')

        return icon_name

    def on_download_icon_complete(self, icon_name, host):
        """
        Runs any download cleanup functions

        :param icon_name: the filename of the icon that finished downloading
        :type icon_name: string
        :param host: the host the icon completed to download for
        :type host: string
        :returns: the icon that finished downloading
        :rtype: TrackerIcon
        """
        log.debug('Successfully downloaded from %s: %s', host, icon_name)
        return TrackerIcon(icon_name)

    def on_download_icon_fail(self, f, host, icons=None):
        """
        Recovers from a download error

        :param f: the failure that occurred
        :type f: Failure
        :param host: the host the icon failed to download for
        :type host: string
        :param icons: the (optional) list of remaining icons
        :type icons: list
        :returns: a Deferred if recovery was possible
                  else the original failure
        :rtype: Deferred or Failure
        """
        if not icons:
            icons = []
        error_msg = f.getErrorMessage()
        log.debug('Error downloading icon from %s: %s', host, error_msg)
        d = f
        if f.check(NoResource, ForbiddenResource) and icons:
            d = self.download_icon(icons, host)
        elif f.check(NoIconsError):
            # No icons, try favicon.ico as an act of desperation
            d = self.download_icon(
                [
                    (
                        urljoin(self.host_to_url(host), 'favicon.ico'),
                        extension_to_mimetype('ico'),
                    )
                ],
                host,
            )
            d.addCallbacks(
                self.on_download_icon_complete,
                self.on_download_icon_fail,
                callbackArgs=(host,),
                errbackArgs=(host,),
            )
        else:
            # No icons :(
            # Return the None Icon
            d = self.icons[None]

        return d

    @proxy(threads.deferToThread)
    def resize_icon(self, icon):
        """
        Resizes the given icon to be 16x16 pixels

        :param icon: the icon to resize
        :type icon: TrackerIcon
        :returns: the resized icon
        :rtype: TrackerIcon
        """
        # Requires Pillow(PIL) to resize.
        if icon and Image:
            filename = icon.get_filename()
            remove_old = False
            with Image.open(filename) as img:
                if img.size > (16, 16):
                    new_filename = filename.rpartition('.')[0] + '.png'
                    img = img.resize((16, 16), Image.ANTIALIAS)
                    img.save(new_filename)
                    if new_filename != filename:
                        remove_old = True
            if remove_old:
                os.remove(filename)
                icon = TrackerIcon(new_filename)
        return icon

    def store_icon(self, icon, host):
        """
        Stores the icon for the given host
        Callbacks any pending deferreds waiting on this icon

        :param icon: the icon to store
        :type icon: TrackerIcon or None
        :param host: the host to store it for
        :type host: string
        :returns: the stored icon
        :rtype: TrackerIcon or None
        """
        self.icons[host] = icon
        for d in self.pending[host]:
            d.callback(icon)
        del self.pending[host]
        return icon

    def host_to_url(self, host):
        """
        Given a host, returns the URL to fetch

        :param host: the tracker host
        :type host: string
        :returns: the url of the tracker
        :rtype: string
        """
        if host in self.redirects:
            host = self.redirects[host]
        return 'http://%s/' % host


# ------- HELPER CLASSES ------


class FaviconParser(HTMLParser):
    """
    A HTMLParser which extracts favicons from a HTML page
    """

    def __init__(self):
        self.icons = []
        self.left_head = False
        HTMLParser.__init__(self)

    def handle_starttag(self, tag, attrs):
        if (
            tag == 'link'
            and ('rel', 'icon') in attrs
            or ('rel', 'shortcut icon') in attrs
        ):
            href = None
            icon_type = None
            for attr, value in attrs:
                if attr == 'href':
                    href = value
                elif attr == 'type':
                    icon_type = value
            if href:
                try:
                    mimetype = extension_to_mimetype(href.rpartition('.')[2])
                except KeyError:
                    pass
                else:
                    icon_type = mimetype
                if icon_type:
                    self.icons.append((href, icon_type))

    def handle_endtag(self, tag):
        if tag == 'head':
            self.left_head = True

    def get_icons(self):
        """
        Returns a list of favicons extracted from the HTML page

        :returns: a list of favicons
        :rtype: list
        """
        return self.icons


# ------ HELPER FUNCTIONS ------


def url_to_host(url):
    """
    Given a URL, returns the host it belongs to

    :param url: the URL in question
    :type url: string
    :returns: the host of the given URL
    :rtype: string
    """
    return urlparse(url).hostname


def host_to_icon_name(host, mimetype):
    """
    Given a host, returns the appropriate icon name

    :param host: the host in question
    :type host: string
    :param mimetype: the mimetype of the icon
    :type mimetype: string
    :returns: the icon's filename
    :rtype: string

    """
    return host + '.' + mimetype_to_extension(mimetype)


def icon_name_to_host(icon):
    """
    Given a host's icon name, returns the host name

    :param icon: the icon name
    :type icon: string
    :returns: the host name
    :rtype: string
    """
    return icon.rpartition('.')[0]


MIME_MAP = {
    'image/gif': 'gif',
    'image/jpeg': 'jpg',
    'image/png': 'png',
    'image/vnd.microsoft.icon': 'ico',
    'image/x-icon': 'ico',
    'image/svg+xml': 'svg',
    'gif': 'image/gif',
    'jpg': 'image/jpeg',
    'jpeg': 'image/jpeg',
    'png': 'image/png',
    'ico': 'image/vnd.microsoft.icon',
    'svg': 'image/svg+xml',
}


def mimetype_to_extension(mimetype):
    """
    Given a mimetype, returns the appropriate filename extension

    :param mimetype: the mimetype
    :type mimetype: string
    :returns: the filename extension for the given mimetype
    :rtype: string
    :raises KeyError: if given an invalid mimetype
    """
    return MIME_MAP[mimetype.lower()]


def extension_to_mimetype(extension):
    """
    Given a filename extension, returns the appropriate mimetype

    :param extension: the filename extension
    :type extension: string
    :returns: the mimetype for the given filename extension
    :rtype: string
    :raises KeyError: if given an invalid filename extension
    """
    return MIME_MAP[extension.lower()]


#  ------ EXCEPTIONS ------


class NoIconsError(Exception):
    pass


class InvalidIconError(Exception):
    pass



#
# Copyright (C) 2008-2009 Ido Abramovich <ido.deluge@gmail.com>
# Copyright (C) 2009 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import os
from base64 import b64encode
from urllib.parse import urlparse
from urllib.request import url2pathname

from twisted.internet import defer

import deluge.common
import deluge.component as component
from deluge.ui.client import client

from . import BaseCommand


class Command(BaseCommand):
    """Add torrents"""

    def add_arguments(self, parser):
        parser.add_argument(
            '-p', '--path', dest='path', help=_('Download folder for torrent')
        )
        parser.add_argument(
            '-m',
            '--move-path',
            dest='move_completed_path',
            help=_('Move the completed torrent to this folder'),
        )
        parser.add_argument(
            'torrents',
            metavar='<torrent>',
            nargs='+',
            help=_('One or more torrent files, URLs or magnet URIs'),
        )

    def handle(self, options):
        self.console = component.get('ConsoleUI')

        t_options = {}
        if options.path:
            t_options['download_location'] = os.path.abspath(
                os.path.expanduser(options.path)
            )

        if options.move_completed_path:
            t_options['move_completed'] = True
            t_options['move_completed_path'] = os.path.abspath(
                os.path.expanduser(options.move_completed_path)
            )

        def on_success(result):
            if not result:
                self.console.write('{!error!}Torrent was not added: Already in session')
            else:
                self.console.write('{!success!}Torrent added!')

        def on_fail(result):
            self.console.write('{!error!}Torrent was not added: %s' % result)

        # Keep a list of deferreds to make a DeferredList
        deferreds = []
        for torrent in options.torrents:
            if not torrent.strip():
                continue
            if deluge.common.is_url(torrent):
                self.console.write(
                    '{!info!}Attempting to add torrent from URL: %s' % torrent
                )
                deferreds.append(
                    client.core.add_torrent_url(torrent, t_options)
                    .addCallback(on_success)
                    .addErrback(on_fail)
                )
            elif deluge.common.is_magnet(torrent):
                self.console.write(
                    '{!info!}Attempting to add torrent from magnet URI: %s' % torrent
                )
                deferreds.append(
                    client.core.add_torrent_magnet(torrent, t_options)
                    .addCallback(on_success)
                    .addErrback(on_fail)
                )
            else:
                # Just a file
                if urlparse(torrent).scheme == 'file':
                    torrent = url2pathname(urlparse(torrent).path)
                path = os.path.abspath(os.path.expanduser(torrent))
                if not os.path.exists(path):
                    self.console.write('{!error!}%s does not exist!' % path)
                    continue
                if not os.path.isfile(path):
                    self.console.write('{!error!}This is a directory!')
                    continue
                self.console.write('{!info!}Attempting to add torrent: %s' % path)
                filename = os.path.split(path)[-1]
                with open(path, 'rb') as _file:
                    filedump = b64encode(_file.read())
                deferreds.append(
                    client.core.add_torrent_file_async(filename, filedump, t_options)
                    .addCallback(on_success)
                    .addErrback(on_fail)
                )

        return defer.DeferredList(deferreds)

    def complete(self, line):
        return component.get('ConsoleUI').tab_complete_path(
            line, ext='.torrent', sort='date'
        )



#
# Copyright (C) 2009-2010 Pedro Algarvio <pedro@algarvio.me>
#
# Basic plugin template created by:
# Copyright (C) 2008 Martijn Voncken <mvoncken@gmail.com>
# Copyright (C) 2007-2009 Andrew Resch <andrewresch@gmail.com>
# Copyright (C) 2009 Damien Churchill <damoxc@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging
import os.path

from pkg_resources import resource_filename
from twisted.internet import defer

from deluge import component
from deluge.event import known_events

log = logging.getLogger(__name__)


def get_resource(filename):
    return resource_filename(__package__, os.path.join('data', filename))


class CustomNotifications:
    def __init__(self, plugin_name=None):
        self.custom_notifications = {'email': {}, 'popup': {}, 'blink': {}, 'sound': {}}

    def enable(self):
        pass

    def disable(self):
        for kind in self.custom_notifications:
            for eventtype in list(self.custom_notifications[kind]):
                wrapper, handler = self.custom_notifications[kind][eventtype]
                self._deregister_custom_provider(kind, eventtype)

    def _handle_custom_providers(self, kind, eventtype, *args, **kwargs):
        log.debug(
            'Calling CORE custom %s providers for %s: %s %s',
            kind,
            eventtype,
            args,
            kwargs,
        )
        if eventtype in self.config['subscriptions'][kind]:
            wrapper, handler = self.custom_notifications[kind][eventtype]
            log.debug('Found handler for kind %s: %s', kind, handler)
            custom_notif_func = getattr(self, 'handle_custom_%s_notification' % kind)
            d = defer.maybeDeferred(handler, *args, **kwargs)
            d.addCallback(custom_notif_func, eventtype)
            d.addCallback(self._on_notify_sucess, kind)
            d.addErrback(self._on_notify_failure, kind)
            return d

    def _register_custom_provider(self, kind, eventtype, handler):
        if not self._handled_eventtype(eventtype, handler):
            return defer.succeed('Event not handled')
        if eventtype not in self.custom_notifications:

            def wrapper(*args, **kwargs):
                return self._handle_custom_providers(kind, eventtype, *args, **kwargs)

            self.custom_notifications[kind][eventtype] = (wrapper, handler)
        else:
            wrapper, handler = self.custom_notifications[kind][eventtype]
        try:
            component.get('EventManager').register_event_handler(eventtype, wrapper)
        except KeyError:
            from deluge.ui.client import client

            client.register_event_handler(eventtype, wrapper)

    def _deregister_custom_provider(self, kind, eventtype):
        try:
            wrapper, handler = self.custom_notifications[kind][eventtype]
            try:
                component.get('EventManager').deregister_event_handler(
                    eventtype, wrapper
                )
            except KeyError:
                from deluge.ui.client import client

                client.deregister_event_handler(eventtype, wrapper)
            self.custom_notifications[kind].pop(eventtype)
        except KeyError:
            pass

    def _handled_eventtype(self, eventtype, handler):
        if eventtype not in known_events:
            log.error('The event "%s" is not known', eventtype)
            return False
        if known_events[eventtype].__module__.startswith('deluge.event'):
            if handler.__self__ is self:
                return True
            log.error(
                'You cannot register custom notification providers '
                'for built-in event types.'
            )
            return False
        return True

    def _on_notify_sucess(self, result, kind):
        log.debug('Notification success using %s: %s', kind, result)
        return result

    def _on_notify_failure(self, failure, kind):
        log.debug('Notification failure using %s: %s', kind, failure)
        return failure



#
# Copyright (C) 2007-2008 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging
import os

from gi.repository import Gtk

import deluge.common
import deluge.component as component
from deluge.ui.client import client

log = logging.getLogger(__name__)


class RemoveTorrentDialog:
    """
    This class is used to create and show a Remove Torrent Dialog.

    :param torrent_ids: the torrent_ids to remove
    :type torrent_ids: list of torrent_ids

    :raises TypeError: if `torrent_id` is not a sequence type
    :raises ValueError: if `torrent_id` contains no torrent_ids or is None

    """

    def __init__(self, torrent_ids, delete_files=False):
        if not isinstance(torrent_ids, list) and not isinstance(torrent_ids, tuple):
            raise TypeError('requires a list of torrent_ids')

        if not torrent_ids:
            raise ValueError('requires a list of torrent_ids')

        self.__torrent_ids = torrent_ids

        self.builder = Gtk.Builder()
        self.builder.add_from_file(
            deluge.common.resource_filename(
                __package__, os.path.join('glade', 'remove_torrent_dialog.ui')
            )
        )

        self.__dialog = self.builder.get_object('remove_torrent_dialog')
        self.__dialog.set_transient_for(component.get('MainWindow').window)

        self.builder.connect_signals(self)
        self.builder.get_object('delete_files').set_active(delete_files)
        label_title = self.builder.get_object('label_title')
        label_torrents = self.builder.get_object('label_torrents')
        num_torrents = len(self.__torrent_ids)
        if num_torrents == 1:
            label_torrents.set_markup(
                component.get('TorrentView').get_torrent_status(self.__torrent_ids[0])[
                    'name'
                ]
            )
        else:
            label_title.set_markup(_('Remove the selected torrents?'))
            label_torrents.set_markup(_('Total of %s torrents selected') % num_torrents)

    def on_delete_files_toggled(self, widget):
        self.builder.get_object('warning_label').set_visible(widget.get_active())

    def __remove_torrents(self, remove_data):
        # Unselect all to avoid issues with the selection changed event
        component.get('TorrentView').treeview.get_selection().unselect_all()

        def on_removed_finished(errors):
            if errors:
                log.info('Error(s) occured when trying to delete torrent(s).')
                for t_id, e_msg in errors:
                    log.warning('Error removing torrent %s : %s', t_id, e_msg)

        d = client.core.remove_torrents(self.__torrent_ids, remove_data)
        d.addCallback(on_removed_finished)

    def run(self):
        """
        Shows the dialog and awaits for user input.  The user can select to
        remove the torrent(s) from the session with or without their data.
        """
        if self.__dialog.run() == Gtk.ResponseType.OK:
            self.__remove_torrents(self.builder.get_object('delete_files').get_active())
        self.__dialog.destroy()
